import argparse
import random
from pathlib import Path
from typing import Any
from typing import get_args

import matplotlib.pyplot as plt
import numpy as np
import torchvision.transforms.v2.functional as F
from torch.utils.data import DataLoader
from torchvision import tv_tensors
from torchvision.utils import draw_bounding_boxes

from birder.common import cli
from birder.common import fs_ops
from birder.common import lib
from birder.conf import settings
from birder.data.collators.detection import collate_fn
from birder.data.datasets.coco import CocoInference
from birder.data.datasets.coco import CocoTraining
from birder.data.datasets.directory import tv_loader
from birder.data.transforms.classification import get_rgb_stats
from birder.data.transforms.classification import reverse_preset
from birder.data.transforms.detection import AugType
from birder.data.transforms.detection import InferenceTransform
from birder.data.transforms.detection import training_preset


# pylint: disable=too-many-locals,too-many-branches
def show_det_iterator(args: argparse.Namespace) -> None:
    reverse_transform = reverse_preset(get_rgb_stats("birder"))
    root_path = Path(args.data_path)
    if args.mode == "training":
        offset = 0
        transform = training_preset(
            args.size,
            args.aug_type,
            args.aug_level,
            get_rgb_stats("birder"),
            args.dynamic_size,
            args.multiscale,
            args.max_size,
        )
        dataset = CocoTraining(args.data_path, args.coco_json_path, transforms=transform)
    elif args.mode == "inference":
        offset = 1
        transform = InferenceTransform(args.size, get_rgb_stats("birder"), args.dynamic_size, args.max_size)
        dataset = CocoInference(args.data_path, args.coco_json_path, transforms=transform)
    else:
        raise ValueError(f"Unknown mode={args.mode}")

    if args.class_file is not None:
        class_to_idx = fs_ops.read_class_file(args.class_file)
        class_to_idx = lib.detection_class_to_idx(class_to_idx)
    else:
        class_to_idx = lib.class_to_idx_from_coco(dataset.dataset.coco.cats)

    class_list = list(class_to_idx.keys())
    class_list.insert(0, "Background")
    color_list = np.arange(0, len(class_list))

    batch_size = 2
    no_iterations = 6
    if args.batch is False:
        samples = random.sample(dataset.dataset.ids, no_iterations)
        cols = 2
        rows = 2
        for coco_id in samples:
            img_path = dataset.dataset.coco.loadImgs(coco_id)[0]["file_name"]
            img = tv_loader(str(root_path.joinpath(img_path)))
            targets = dataset.dataset.coco.loadAnns(dataset.dataset.coco.getAnnIds(coco_id))

            # Roughly what the "wrap_dataset_for_transforms_v2" does
            canvas_size = tuple(F.get_size(img))
            boxes = F.convert_bounding_box_format(
                tv_tensors.BoundingBoxes(
                    [target["bbox"] for target in targets],
                    format=tv_tensors.BoundingBoxFormat.XYWH,
                    canvas_size=canvas_size,
                ),
                new_format=tv_tensors.BoundingBoxFormat.XYXY,
            )

            label_ids = [target["category_id"] for target in targets]
            labels = [class_list[label_id] for label_id in label_ids]
            colors = [color_list[label_id].item() for label_id in label_ids]

            annotated_img = draw_bounding_boxes(img, boxes, labels=labels, colors=colors)
            transformed_img = F.to_pil_image(annotated_img)

            fig = plt.figure(constrained_layout=True)
            grid_spec = fig.add_gridspec(ncols=cols, nrows=rows)

            # Show original
            ax = fig.add_subplot(grid_spec[0, 0:cols])
            ax.imshow(transformed_img)
            ax.set_title(f"Original, aug type: {args.aug_type}")

            # Show transformed
            counter = 0
            for i in range(cols):
                for j in range(1, rows):
                    idx = dataset.dataset.ids.index(coco_id)
                    data = dataset[idx]
                    img = data[offset]
                    targets = data[offset + 1]
                    label_ids = targets["labels"]
                    labels = [class_list[label_id] for label_id in label_ids]
                    colors = [color_list[label_id].item() for label_id in label_ids]

                    annotated_img = draw_bounding_boxes(
                        reverse_transform(img), targets["boxes"], labels=labels, colors=colors
                    )
                    transformed_img = F.to_pil_image(annotated_img)

                    ax = fig.add_subplot(grid_spec[j, i])
                    ax.imshow(np.asarray(transformed_img))
                    ax.set_title(f"#{counter}")
                    counter += 1

            plt.show()

    else:
        data_loader = DataLoader(
            dataset,
            batch_size=batch_size,
            shuffle=True,
            collate_fn=collate_fn,
        )

        cols = 2
        rows = 1
        for k, data in enumerate(data_loader):
            if k >= no_iterations:
                break

            inputs = data[offset]
            targets = data[offset + 1]
            if offset > 0:
                sample_paths = data[offset - 1]
            else:
                sample_paths = list(range(cols * rows))

            fig = plt.figure(constrained_layout=True)
            grid_spec = fig.add_gridspec(ncols=cols, nrows=rows)

            # Show transformed
            counter = 0
            for i in range(cols):
                for j in range(rows):
                    img = inputs[i + cols * j]
                    img = reverse_transform(img)
                    boxes = targets[i + cols * j]["boxes"]
                    label_ids = targets[i + cols * j]["labels"]
                    labels = [class_list[label_id] for label_id in label_ids]
                    colors = [color_list[label_id].item() for label_id in label_ids]

                    annotated_img = draw_bounding_boxes(img, boxes, labels=labels, colors=colors)
                    transformed_img = F.to_pil_image(annotated_img)
                    ax = fig.add_subplot(grid_spec[j, i])
                    ax.imshow(np.asarray(transformed_img))
                    ax.set_title(f"#{counter}: {sample_paths[i+cols*j]}")
                    counter += 1

            plt.show()


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "show-det-iterator",
        allow_abbrev=False,
        help="show training / inference detection iterator output vs input",
        description="show training / inference detection iterator output vs input",
        epilog=(
            "Usage examples:\n"
            "python -m birder.tools show-det-iterator --aug-level 0\n"
            "python -m birder.tools show-det-iterator --mode training --size 512 --aug-level 2\n"
            "python -m birder.tools show-det-iterator --mode inference --size 640\n"
            "python -m birder.tools show-det-iterator --mode inference --coco-json-path "
            "~/Datasets/Objects365-2020/val/zhiyuan_objv2_val.json --data-path ~/Datasets/Objects365-2020/val\n"
            "python -m birder.tools show-det-iterator --aug-type ssd --dynamic-size --coco-json-path "
            "~/Datasets/cocodataset/annotations/instances_val2017.json --data-path "
            "~/Datasets/cocodataset/val2017 --class-file public_datasets_metadata/coco-classes.txt\n"
            "python tool.py show-det-iterator --mode training --aug-level 4 --multiscale "
            "--coco-json-path ~/Datasets/cocodataset/annotations/instances_val2017.json "
            "--data-path ~/Datasets/cocodataset/val2017 --class-file public_datasets_metadata/coco-classes.txt\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument(
        "--mode", type=str, choices=["training", "inference"], default="training", help="iterator mode"
    )
    subparser.add_argument(
        "--batch", default=False, action="store_true", help="show a batch instead of a single sample"
    )
    subparser.add_argument(
        "--size",
        type=int,
        nargs="+",
        default=[512],
        metavar=("H", "W"),
        help=(
            "target image size as [height, width], if --dynamic-size is enabled, "
            "uses the smaller dimension as target size while preserving aspect ratio (defaults to model's signature)"
        ),
    )
    subparser.add_argument(
        "--max-size",
        type=int,
        help="maximum size for the longer edge of resized images, when specified, enables dynamic sizing",
    )
    subparser.add_argument(
        "--dynamic-size",
        default=False,
        action="store_true",
        help="allow variable image sizes while preserving aspect ratios",
    )
    subparser.add_argument("--multiscale", default=False, action="store_true", help="enable random scale per image")
    subparser.add_argument(
        "--aug-type",
        type=str,
        choices=list(get_args(AugType)),
        default="birder",
        help="augmentation type",
    )
    subparser.add_argument(
        "--aug-level",
        type=int,
        choices=list(range(10 + 1)),
        default=4,
        help="magnitude of birder augmentations (0 off -> 10 highest)",
    )
    subparser.add_argument(
        "--data-path",
        type=str,
        default=str(settings.DETECTION_DATA_PATH),
        help="image directory path",
    )
    subparser.add_argument(
        "--coco-json-path",
        type=str,
        default=f"{settings.TRAINING_DETECTION_ANNOTATIONS_PATH}_coco.json",
        help="training COCO json path",
    )
    subparser.add_argument("--class-file", type=str, metavar="FILE", help="class list file, overrides json categories")
    subparser.set_defaults(func=main)


def main(args: argparse.Namespace) -> None:
    args.size = cli.parse_size(args.size)

    if args.multiscale is True and args.aug_type != "birder":
        raise cli.ValidationError(f"--multiscale only supported with --aug-type birder, got {args.aug_type}")

    show_det_iterator(args)



"""
CAS-ViT, adapted from
https://github.com/Tianfang-Zhang/CAS-ViT/blob/main/classification/model/rcvit.py

Paper "CAS-ViT: Convolutional Additive Self-attention Vision Transformers for Efficient Mobile Applications",
https://arxiv.org/abs/2408.03703

Changes from original:
* Removed biases before norms
"""

# Reference license: MIT

from collections import OrderedDict
from collections.abc import Callable
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import StochasticDepth

from birder.model_registry import registry
from birder.net.base import DetectorBackbone


class Downsample(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        padding: tuple[int, int],
    ) -> None:
        super().__init__()
        self.proj = nn.Conv2d(
            in_channels, out_channels, kernel_size=kernel_size, stride=stride, padding=padding, bias=False
        )
        self.norm = nn.BatchNorm2d(out_channels)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.proj(x)
        x = self.norm(x)
        return x


class ConvMLP(nn.Module):
    def __init__(
        self,
        in_features: int,
        hidden_features: int,
        out_features: int,
        act_layer: Callable[..., nn.Module] = nn.GELU,
    ) -> None:
        super().__init__()
        self.fc1 = nn.Conv2d(in_features, hidden_features, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.act = act_layer()
        self.fc2 = nn.Conv2d(hidden_features, out_features, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.fc1(x)
        x = self.act(x)
        x = self.fc2(x)
        return x


class SpatialOperation(nn.Module):
    def __init__(self, dim: int) -> None:
        super().__init__()
        self.block = nn.Sequential(
            nn.Conv2d(dim, dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=dim),
            nn.BatchNorm2d(dim),
            nn.ReLU(),
            nn.Conv2d(dim, 1, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False),
            nn.Sigmoid(),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x * self.block(x)


class ChannelOperation(nn.Module):
    def __init__(self, dim: int) -> None:
        super().__init__()
        self.block = nn.Sequential(
            nn.AdaptiveAvgPool2d((1, 1)),
            nn.Conv2d(dim, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False),
            nn.Sigmoid(),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x * self.block(x)


class LocalIntegration(nn.Module):
    def __init__(self, dim: int, ratio: float, act_layer: Callable[..., nn.Module] = nn.GELU) -> None:
        super().__init__()
        mid_dim = round(ratio * dim)
        self.block = nn.Sequential(
            nn.Conv2d(dim, mid_dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False),
            nn.BatchNorm2d(mid_dim),
            nn.Conv2d(mid_dim, mid_dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=mid_dim),
            act_layer(),
            nn.Conv2d(mid_dim, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0)),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.block(x)


class AdditiveTokenMixer(nn.Module):
    def __init__(self, dim: int, qkv_bias: bool, proj_drop: float) -> None:
        super().__init__()
        self.qkv = nn.Conv2d(dim, 3 * dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=qkv_bias)
        self.op_q = nn.Sequential(SpatialOperation(dim), ChannelOperation(dim))
        self.op_k = nn.Sequential(SpatialOperation(dim), ChannelOperation(dim))
        self.dwc = nn.Conv2d(dim, dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=dim)

        self.proj = nn.Conv2d(dim, dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (q, k, v) = self.qkv(x).chunk(3, dim=1)
        q = self.op_q(q)
        k = self.op_k(k)

        x = self.proj(self.dwc(q + k) * v)
        x = self.proj_drop(x)

        return x


class AdditiveBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        mlp_ratio: float,
        qkv_bias: bool,
        proj_drop: float,
        drop_path: float,
        act_layer: Callable[..., nn.Module] = nn.GELU,
    ) -> None:
        super().__init__()
        self.local_perception = LocalIntegration(dim, ratio=1.0, act_layer=act_layer)
        self.norm1 = nn.BatchNorm2d(dim)
        self.attn = AdditiveTokenMixer(dim, qkv_bias=qkv_bias, proj_drop=proj_drop)

        self.norm2 = nn.BatchNorm2d(dim)
        self.mlp = ConvMLP(dim, hidden_features=int(dim * mlp_ratio), out_features=dim, act_layer=act_layer)

        self.drop_path = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.local_perception(x)
        x = x + self.drop_path(self.attn(self.norm1(x)))
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x


# pylint: disable=invalid-name
class CAS_ViTStage(nn.Module):
    def __init__(
        self,
        dim: int,
        out_dim: int,
        mlp_ratio: float,
        qkv_bias: bool,
        proj_drop: float,
        drop_path: list[float],
        depth: int,
        downsample: bool,
    ) -> None:
        super().__init__()
        if downsample is True:
            self.downsample = Downsample(dim, out_dim, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1))
        else:
            assert dim == out_dim
            self.downsample = nn.Identity()

        layers = []
        for i in range(depth):
            layers.append(
                AdditiveBlock(
                    out_dim, mlp_ratio=mlp_ratio, qkv_bias=qkv_bias, proj_drop=proj_drop, drop_path=drop_path[i]
                )
            )

        self.blocks = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.downsample(x)
        x = self.blocks(x)

        return x


# pylint: disable=invalid-name
class CAS_ViT(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        mlp_ratio = 4.0
        qkv_bias = False
        proj_drop = 0.0
        depths: list[int] = self.config["depths"]
        embed_dims: list[int] = self.config["embed_dims"]
        drop_path_rate: float = self.config["drop_path_rate"]

        self.stem = nn.Sequential(
            Conv2dNormActivation(
                self.input_channels, embed_dims[0] // 2, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)
            ),
            Conv2dNormActivation(embed_dims[0] // 2, embed_dims[0], kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
        )

        dpr = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(depths)).split(depths)]
        num_stages = len(depths)

        prev_dim = embed_dims[0]
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i in range(num_stages):
            stages[f"stage{i+1}"] = CAS_ViTStage(
                prev_dim,
                embed_dims[i],
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                proj_drop=proj_drop,
                drop_path=dpr[i],
                depth=depths[i],
                downsample=i > 0,
            )
            return_channels.append(embed_dims[i])
            prev_dim = embed_dims[i]

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.BatchNorm2d(embed_dims[-1]),
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = embed_dims[-1]
        self.dist_classifier = self.create_classifier()
        self.classifier = self.create_classifier()
        self.distillation_output = False

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    def reset_classifier(self, num_classes: int) -> None:
        self.num_classes = num_classes
        self.dist_classifier = self.create_classifier()
        self.classifier = self.create_classifier()

    def freeze(self, freeze_classifier: bool = True, unfreeze_features: bool = False) -> None:
        for param in self.parameters():
            param.requires_grad = False

        if freeze_classifier is False:
            for param in self.classifier.parameters():
                param.requires_grad = True

            for param in self.dist_classifier.parameters():
                param.requires_grad = True

        if unfreeze_features is True:
            for param in self.features.parameters():
                param.requires_grad = True

    def transform_to_backbone(self) -> None:
        self.features = nn.Identity()
        self.classifier = nn.Identity()
        self.dist_classifier = nn.Identity()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def set_distillation_output(self, enable: bool = True) -> None:
        self.distillation_output = enable

    def classify(self, x: torch.Tensor) -> torch.Tensor:
        x_cls = self.classifier(x)
        x_dist = self.dist_classifier(x)

        if self.training is True and self.distillation_output is True:
            x = torch.stack([x_cls, x_dist], dim=1)
        else:
            # Classifier "token" as an average of both tokens (during normal training or inference)
            x = (x_cls + x_dist) / 2

        return x


registry.register_model_config(
    "cas_vit_xs", CAS_ViT, config={"depths": [2, 2, 4, 2], "embed_dims": [48, 56, 112, 220], "drop_path_rate": 0.0}
)
registry.register_model_config(
    "cas_vit_s", CAS_ViT, config={"depths": [3, 3, 6, 3], "embed_dims": [48, 64, 128, 256], "drop_path_rate": 0.0}
)
registry.register_model_config(
    "cas_vit_m", CAS_ViT, config={"depths": [3, 3, 6, 3], "embed_dims": [64, 96, 192, 384], "drop_path_rate": 0.0}
)
registry.register_model_config(
    "cas_vit_t", CAS_ViT, config={"depths": [3, 3, 6, 3], "embed_dims": [96, 128, 256, 512], "drop_path_rate": 0.0}
)

registry.register_weights(
    "cas_vit_xs_il-common",
    {
        "description": "CAS-ViT extra small model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 11.4,
                "sha256": "e8d1c09f98a1a33334e7281d471684d260b312ac02d467bb6740e4a98d00daa8",
            }
        },
        "net": {"network": "cas_vit_xs", "tag": "il-common"},
    },
)



"""
BiFormer, adapted from
https://github.com/rayleizhu/BiFormer/blob/public_release/models/biformer.py

"BiFormer: Vision Transformer with Bi-Level Routing Attention", https://arxiv.org/abs/2303.08810

Changes from original:
* All attention types are in (B, C, H, W)
* Using the newer Bi-Level Routing Attention implementation
* Dynamic n_win size (image size // 32)
"""

# Reference license: Apache-2.0

from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import StochasticDepth

from birder.layers import LayerNorm2d
from birder.layers import LayerScale2d
from birder.model_registry import registry
from birder.net.base import DetectorBackbone


def _grid2seq(x: torch.Tensor, region_size: tuple[int, int], num_heads: int) -> tuple[torch.Tensor, int, int]:
    (B, C, H, W) = x.size()
    region_h = H // region_size[0]
    region_w = W // region_size[1]
    x = x.view(B, num_heads, C // num_heads, region_h, region_size[0], region_w, region_size[1])
    x = torch.einsum("bmdhpwq->bmhwpqd", x).flatten(2, 3).flatten(-3, -2)  # (bs, n_head, n_region, reg_size, head_dim)

    return (x, region_h, region_w)


def _seq2grid(x: torch.Tensor, region_h: int, region_w: int, region_size: tuple[int, int]) -> torch.Tensor:
    (bs, n_head, _, _, head_dim) = x.size()
    x = x.view(bs, n_head, region_h, region_w, region_size[0], region_size[1], head_dim)
    x = torch.einsum("bmhwpqd->bmdhpwq", x).reshape(
        bs, n_head * head_dim, region_h * region_size[0], region_w * region_size[1]
    )

    return x


# pylint: disable=too-many-locals
def regional_routing_attention_torch(
    query: torch.Tensor,
    key: torch.Tensor,
    value: torch.Tensor,
    scale: float,
    region_graph: torch.Tensor,  # LongTensor
    region_size: tuple[int, int],
    auto_pad: bool,
) -> tuple[torch.Tensor, torch.Tensor]:
    kv_region_size = region_size
    (bs, n_head, q_nregion, topk) = region_graph.size()

    # Pad to deal with any input size
    q_pad_b = 0
    q_pad_r = 0
    kv_pad_b = 0
    kv_pad_r = 0
    if auto_pad is True:
        (_, _, h_q, w_q) = query.size()
        q_pad_b = (region_size[0] - h_q % region_size[0]) % region_size[0]
        q_pad_r = (region_size[1] - w_q % region_size[1]) % region_size[1]
        if q_pad_b > 0 or q_pad_r > 0:
            query = F.pad(query, (0, q_pad_r, 0, q_pad_b))

        (_, _, h_k, w_k) = key.size()
        kv_pad_b = (kv_region_size[0] - h_k % kv_region_size[0]) % kv_region_size[0]
        kv_pad_r = (kv_region_size[1] - w_k % kv_region_size[1]) % kv_region_size[1]
        if kv_pad_r > 0 or kv_pad_b > 0:
            key = F.pad(key, (0, kv_pad_r, 0, kv_pad_b))
            value = F.pad(value, (0, kv_pad_r, 0, kv_pad_b))
    else:
        h_q = None
        w_q = None
        h_k = None
        w_k = None

    # To sequence format
    (query, q_region_h, q_region_w) = _grid2seq(query, region_size=region_size, num_heads=n_head)
    (key, _, _) = _grid2seq(key, region_size=kv_region_size, num_heads=n_head)
    (value, _, _) = _grid2seq(value, region_size=kv_region_size, num_heads=n_head)

    # Gather key and values
    (bs, n_head, kv_nregion, kv_region_size, head_dim) = key.size()
    broadcasted_region_graph = region_graph.view(bs, n_head, q_nregion, topk, 1, 1).expand(
        -1, -1, -1, -1, kv_region_size, head_dim
    )
    key_g = torch.gather(
        key.view(bs, n_head, 1, kv_nregion, kv_region_size, head_dim).expand(-1, -1, query.size(2), -1, -1, -1),
        dim=3,
        index=broadcasted_region_graph,
    )  # (bs, n_head, q_nregion, topk, kv_region_size, head_dim)
    value_g = torch.gather(
        value.view(bs, n_head, 1, kv_nregion, kv_region_size, head_dim).expand(-1, -1, query.size(2), -1, -1, -1),
        dim=3,
        index=broadcasted_region_graph,
    )  # (bs, n_head, q_nregion, topk, kv_region_size, head_dim)

    # Token-to-token attention
    attn = (query * scale) @ key_g.flatten(-3, -2).transpose(-1, -2)
    attn = torch.softmax(attn, dim=-1)
    output = attn @ value_g.flatten(-3, -2)

    # To grid format
    output = _seq2grid(output, region_h=q_region_h, region_w=q_region_w, region_size=region_size)

    # Remove paddings if needed
    if auto_pad and (q_pad_b > 0 or q_pad_r > 0):
        output = output[:, :, :h_q, :w_q]

    return (output, attn)


class BiLevelRoutingAttention(nn.Module):
    def __init__(self, dim: int, num_heads: int, n_win_h: int, n_win_w: int, topk: int, side_dwconv: int) -> None:
        super().__init__()
        self.num_heads = num_heads
        self.head_dim = dim // self.num_heads
        self.scale = dim**-0.5
        assert dim % num_heads == 0

        self.lepe = nn.Conv2d(
            dim,
            dim,
            kernel_size=(side_dwconv, side_dwconv),
            stride=(1, 1),
            padding=(side_dwconv // 2, side_dwconv // 2),
            groups=dim,
        )
        self.topk = topk
        self.n_win_h = n_win_h
        self.n_win_w = n_win_w

        self.qkv_linear = nn.Conv2d(dim, 3 * dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.output_linear = nn.Conv2d(dim, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (_, _, H, W) = x.size()
        region_size = (H // self.n_win_h, W // self.n_win_w)

        # Linear projection
        qkv = self.qkv_linear(x)
        (q, k, v) = qkv.chunk(3, dim=1)

        # Region-to-region routing
        q_r = F.avg_pool2d(  # pylint: disable=not-callable
            q.detach(), kernel_size=region_size, stride=region_size, ceil_mode=True, count_include_pad=False
        )
        k_r = F.avg_pool2d(  # pylint: disable=not-callable
            k.detach(), kernel_size=region_size, stride=region_size, ceil_mode=True, count_include_pad=False
        )
        q_r = q_r.permute(0, 2, 3, 1).flatten(1, 2)  # (n, (hw), c)
        k_r = k_r.flatten(2, 3)  # (n, c, (hw))
        a_r = q_r @ k_r
        (_, idx_r) = torch.topk(a_r, k=self.topk, dim=-1)
        idx_r = idx_r.unsqueeze_(1).expand(-1, self.num_heads, -1, -1)

        # Token to token attention
        (output, _) = regional_routing_attention_torch(
            q, k, v, scale=self.scale, region_graph=idx_r, region_size=region_size, auto_pad=True
        )

        output = output + self.lepe(v)
        output = self.output_linear(output)

        return output


class Attention(nn.Module):
    def __init__(self, dim: int, num_heads: int, qkv_bias: bool, attn_drop: float, proj_drop: float) -> None:
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, C, H, W) = x.size()
        x = x.permute(0, 2, 3, 1).reshape(B, H * W, C)

        N = H * W
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        (q, k, v) = qkv.unbind(0)

        x = F.scaled_dot_product_attention(  # pylint: disable=not-callable
            q, k, v, dropout_p=self.attn_drop.p if self.training else 0.0, scale=self.scale
        )
        x = x.transpose(1, 2).reshape(B, N, C)

        x = self.proj(x)
        x = self.proj_drop(x)

        x = x.reshape(B, C, H, W).permute(0, 3, 1, 2)

        return x


class AttentionLePE(nn.Module):
    """
    Attention with Locally-enhanced Positional Encoding
    """

    def __init__(
        self, dim: int, num_heads: int, qkv_bias: bool, side_dwconv: int, attn_drop: float, proj_drop: float
    ) -> None:
        super().__init__()
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.scale = self.head_dim**-0.5

        self.qkv = nn.Conv2d(dim, dim * 3, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Conv2d(dim, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.proj_drop = nn.Dropout(proj_drop)
        self.lepe = nn.Conv2d(
            dim,
            dim,
            kernel_size=(side_dwconv, side_dwconv),
            stride=(1, 1),
            padding=(side_dwconv // 2, side_dwconv // 2),
            groups=dim,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, C, H, W) = x.size()
        (q, k, v) = self.qkv(x).chunk(3, dim=1)

        attn = q.view(B, self.num_heads, self.head_dim, H * W).transpose(-1, -2) @ k.view(
            B, self.num_heads, self.head_dim, H * W
        )
        attn = torch.softmax(attn * self.scale, dim=-1)
        attn = self.attn_drop(attn)

        output = attn @ v.view(B, self.num_heads, self.head_dim, H * W).transpose(-1, -2)
        output = output.permute(0, 1, 3, 2).reshape(B, C, H, W)
        output = output + self.lepe(v)

        output = self.proj(output)
        output = self.proj_drop(output)

        return output


class BiFormerBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        n_win_h: int,
        n_win_w: int,
        topk: int,
        mlp_ratio: float,
        side_dwconv: int,
        layer_scale_init_value: Optional[float],
        drop_path: float,
    ) -> None:
        super().__init__()
        self.pos_embed = nn.Conv2d(dim, dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=dim)
        self.norm1 = LayerNorm2d(dim, eps=1e-6)
        if topk > 0:
            self.attn = BiLevelRoutingAttention(
                dim=dim, num_heads=num_heads, n_win_h=n_win_h, n_win_w=n_win_w, topk=topk, side_dwconv=side_dwconv
            )
        elif topk == -1:
            self.attn = Attention(dim, num_heads=8, qkv_bias=False, attn_drop=0.0, proj_drop=0.0)
        elif topk == -2:
            self.attn = AttentionLePE(
                dim, num_heads=8, qkv_bias=False, side_dwconv=side_dwconv, attn_drop=0.0, proj_drop=0.0
            )
        elif topk == 0:
            self.attn = nn.Sequential(
                nn.Conv2d(dim, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0)),
                nn.Conv2d(dim, dim, kernel_size=(5, 5), stride=(1, 1), padding=(2, 2), groups=dim),
                nn.Conv2d(dim, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0)),
            )
        else:
            raise ValueError(f"topk={topk} not supported")

        self.norm2 = LayerNorm2d(dim, eps=1e-6)
        self.mlp = nn.Sequential(
            nn.Conv2d(dim, int(mlp_ratio * dim), kernel_size=(1, 1), stride=(1, 1), padding=(0, 0)),
            nn.GELU(),
            nn.Conv2d(int(mlp_ratio * dim), dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0)),
        )
        self.drop_path = StochasticDepth(drop_path, mode="row")

        if layer_scale_init_value is not None:
            self.layer_scale_1 = LayerScale2d(dim, layer_scale_init_value)
            self.layer_scale_2 = LayerScale2d(dim, layer_scale_init_value)
        else:
            self.layer_scale_1 = nn.Identity()
            self.layer_scale_2 = nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.pos_embed(x)

        x = x + self.drop_path(self.layer_scale_1(self.attn(self.norm1(x))))
        x = x + self.drop_path(self.layer_scale_2(self.mlp(self.norm2(x))))

        return x


class BiFormerStage(nn.Module):
    def __init__(
        self,
        dim: int,
        out_dim: int,
        num_heads: int,
        n_win_h: int,
        n_win_w: int,
        topk: int,
        mlp_ratio: float,
        side_dwconv: int,
        layer_scale_init_value: Optional[float],
        drop_path: list[float],
        depth: int,
        downsample: bool,
    ) -> None:
        super().__init__()
        if downsample is True:
            self.downsample = nn.Sequential(
                nn.Conv2d(dim, out_dim, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
                nn.BatchNorm2d(out_dim),
            )
        else:
            assert dim == out_dim
            self.downsample = nn.Identity()

        layers = []
        for i in range(depth):
            layers.append(
                BiFormerBlock(
                    out_dim,
                    num_heads=num_heads,
                    n_win_h=n_win_h,
                    n_win_w=n_win_w,
                    topk=topk,
                    mlp_ratio=mlp_ratio,
                    side_dwconv=side_dwconv,
                    layer_scale_init_value=layer_scale_init_value,
                    drop_path=drop_path[i],
                )
            )

        self.blocks = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.downsample(x)
        x = self.blocks(x)

        return x


class BiFormer(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        head_dim = 32
        topks = [1, 4, 16, -2]
        mlp_ratios = [3, 3, 3, 3]
        side_dwconv = 5
        depths: list[int] = self.config["depths"]
        embed_dims: list[int] = self.config["embed_dims"]
        qk_dims: list[int] = self.config["qk_dims"]
        layer_scale_init_value: Optional[float] = self.config["layer_scale_init_value"]
        drop_path_rate: float = self.config["drop_path_rate"]
        n_win_h = self.size[0] // 32
        n_win_w = self.size[1] // 32

        self.stem = nn.Sequential(
            Conv2dNormActivation(
                self.input_channels,
                embed_dims[0] // 2,
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=(1, 1),
                activation_layer=nn.GELU,
                inplace=None,
            ),
            Conv2dNormActivation(
                embed_dims[0] // 2,
                embed_dims[0],
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=(1, 1),
                activation_layer=None,
            ),
        )

        dpr = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(depths)).split(depths)]
        num_stages = len(depths)
        n_heads = [dim // head_dim for dim in qk_dims]

        prev_dim = embed_dims[0]
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i in range(num_stages):
            stages[f"stage{i+1}"] = BiFormerStage(
                prev_dim,
                embed_dims[i],
                num_heads=n_heads[i],
                n_win_h=n_win_h,
                n_win_w=n_win_w,
                topk=topks[i],
                mlp_ratio=mlp_ratios[i],
                side_dwconv=side_dwconv,
                layer_scale_init_value=layer_scale_init_value,
                drop_path=dpr[i],
                depth=depths[i],
                downsample=i > 0,
            )
            return_channels.append(embed_dims[i])
            prev_dim = embed_dims[i]

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.BatchNorm2d(embed_dims[-1]),
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = embed_dims[-1]
        self.classifier = self.create_classifier()

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, nn.LayerNorm):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        assert dynamic_size is False, "Dynamic size not supported for this network"

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        super().adjust_size(new_size)

        n_win_h = new_size[0] // 32
        n_win_w = new_size[1] // 32
        for m in self.modules():
            if isinstance(m, BiLevelRoutingAttention):
                m.n_win_h = n_win_h
                m.n_win_w = n_win_w


registry.register_model_config(
    "biformer_t",
    BiFormer,
    config={
        "depths": [2, 2, 8, 2],
        "embed_dims": [64, 128, 256, 512],
        "qk_dims": [64, 128, 256, 512],
        "layer_scale_init_value": None,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "biformer_s",
    BiFormer,
    config={
        "depths": [4, 4, 18, 4],
        "embed_dims": [64, 128, 256, 512],
        "qk_dims": [64, 128, 256, 512],
        "layer_scale_init_value": None,
        "drop_path_rate": 0.15,
    },
)
registry.register_model_config(
    "biformer_b",
    BiFormer,
    config={
        "depths": [4, 4, 18, 4],
        "embed_dims": [96, 192, 384, 768],
        "qk_dims": [96, 192, 384, 768],
        "layer_scale_init_value": None,
        "drop_path_rate": 0.4,
    },
)

registry.register_weights(
    "biformer_s_il-all",
    {
        "url": "https://huggingface.co/birder-project/biformer_s_il-all/resolve/main",
        "description": "BiFormer small model trained on the il-all dataset",
        "resolution": (384, 384),
        "formats": {
            "pt": {
                "file_size": 96.8,
                "sha256": "ecce546d5adde5cc335a98245b9cc7fddb0ba43628b0d105d974ae6537cbacf7",
            }
        },
        "net": {"network": "biformer_s", "tag": "il-all"},
    },
)



"""
TransNeXt, adapted from
https://github.com/DaiShiResearch/TransNeXt/blob/main/classification/transnext.py

Paper "TransNeXt: Robust Foveal Visual Perception for Vision Transformers", https://arxiv.org/abs/2311.17132
"""

# Reference license: Apache-2.0

import math
from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import StochasticDepth

from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.ops.swattention import SWAttention_AV
from birder.ops.swattention import SWAttention_QK_RPB


@torch.no_grad()  # type: ignore[misc]
def get_relative_position_cpb(
    query_size: tuple[int, int], key_size: tuple[int, int]
) -> tuple[torch.Tensor, torch.Tensor]:
    pretrain_size = query_size
    axis_qh = torch.arange(query_size[0], dtype=torch.float32)
    axis_kh = F.adaptive_avg_pool1d(axis_qh.unsqueeze(0), key_size[0]).squeeze(0)  # pylint: disable=not-callable
    axis_qw = torch.arange(query_size[1], dtype=torch.float32)
    axis_kw = F.adaptive_avg_pool1d(axis_qw.unsqueeze(0), key_size[1]).squeeze(0)  # pylint: disable=not-callable
    (axis_kh, axis_kw) = torch.meshgrid(axis_kh, axis_kw, indexing="ij")
    (axis_qh, axis_qw) = torch.meshgrid(axis_qh, axis_qw, indexing="ij")

    axis_kh = torch.reshape(axis_kh, [-1])
    axis_kw = torch.reshape(axis_kw, [-1])
    axis_qh = torch.reshape(axis_qh, [-1])
    axis_qw = torch.reshape(axis_qw, [-1])

    relative_h = (axis_qh[:, None] - axis_kh[None, :]) / (pretrain_size[0] - 1) * 8
    relative_w = (axis_qw[:, None] - axis_kw[None, :]) / (pretrain_size[1] - 1) * 8
    relative_hw = torch.stack([relative_h, relative_w], dim=-1).view(-1, 2)

    (relative_coords_table, idx_map) = torch.unique(relative_hw, return_inverse=True, dim=0)

    relative_coords_table = (
        torch.sign(relative_coords_table)
        * torch.log2(torch.abs(relative_coords_table) + 1.0)
        / torch.log2(torch.tensor(8, dtype=torch.float32))
    )

    return (idx_map, relative_coords_table)


@torch.no_grad()  # type: ignore[misc]
def get_seqlen_and_mask(input_resolution: tuple[int, int], window_size: int) -> tuple[torch.Tensor, torch.Tensor]:
    attn_map = F.unfold(
        torch.ones([1, 1, input_resolution[0], input_resolution[1]]),
        window_size,
        dilation=1,
        padding=(window_size // 2, window_size // 2),
        stride=1,
    )
    attn_local_length = attn_map.sum(-2).squeeze().unsqueeze(-1)
    attn_mask = (attn_map.squeeze(0).permute(1, 0)) == 0

    return (attn_local_length, attn_mask)


class ConvolutionalGLU(nn.Module):
    def __init__(self, in_features: int, hidden_features: int, drop: float) -> None:
        super().__init__()
        out_features = in_features
        hidden_features = int(2 * hidden_features / 3)
        self.fc1 = nn.Linear(in_features, hidden_features * 2)
        self.dwconv = nn.Conv2d(
            hidden_features, hidden_features, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=hidden_features
        )
        self.act = nn.GELU()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x: torch.Tensor, H: int, W: int) -> torch.Tensor:
        (x, v) = self.fc1(x).chunk(2, dim=-1)

        (B, _, C) = x.size()
        x = x.transpose(1, 2).view(B, C, H, W).contiguous()
        x = self.dwconv(x)
        x = x.flatten(2).transpose(1, 2)

        x = self.act(x) * v
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)

        return x


class Attention(nn.Module):
    def __init__(
        self,
        dim: int,
        input_resolution: tuple[int, int],
        num_heads: int,
        qkv_bias: bool,
        attn_drop: float,
        proj_drop: float,
    ) -> None:
        super().__init__()
        assert dim % num_heads == 0, f"dim {dim} should be divided by num_heads {num_heads}"

        self.dim = dim
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.temperature = nn.Parameter(  # Initialize softplus(temperature) to 1/0.24
            torch.log((torch.ones(num_heads, 1, 1) / 0.24).exp() - 1)
        )

        # Generate sequence length scale
        self.seq_length_scale = nn.Buffer(
            torch.log(torch.tensor(input_resolution[0] * input_resolution[1])), persistent=False
        )

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.query_embedding = nn.Parameter(
            nn.init.trunc_normal_(torch.empty(self.num_heads, 1, self.head_dim), mean=0, std=0.02)
        )

        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        # MLP to generate continuous relative position bias
        self.cpb_fc1 = nn.Linear(2, 512)
        self.cpb_act = nn.ReLU()
        self.cpb_fc2 = nn.Linear(512, num_heads)

    def forward(
        self, x: torch.Tensor, _h: int, _w: int, relative_pos_index: torch.Tensor, relative_coords_table: torch.Tensor
    ) -> torch.Tensor:
        (B, N, C) = x.size()
        qkv = self.qkv(x).reshape(B, -1, 3 * self.num_heads, self.head_dim).permute(0, 2, 1, 3)
        (q, k, v) = qkv.chunk(3, dim=1)

        # Use MLP to generate continuous relative positional bias
        rel_bias = (
            self.cpb_fc2(self.cpb_act(self.cpb_fc1(relative_coords_table)))
            .transpose(0, 1)[:, relative_pos_index.view(-1)]
            .view(-1, N, N)
        )

        # Calculate attention map using sequence length scaled cosine attention and query embedding
        attn = (
            (F.normalize(q, dim=-1) + self.query_embedding)
            * F.softplus(self.temperature)  # pylint: disable=not-callable
            * self.seq_length_scale
        ) @ F.normalize(k, dim=-1).transpose(-2, -1) + rel_bias

        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)
        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


# pylint: disable=too-many-instance-attributes
class AggregatedAttention(nn.Module):
    def __init__(
        self,
        dim: int,
        input_resolution: tuple[int, int],
        num_heads: int,
        window_size: int,
        qkv_bias: bool,
        attn_drop: float,
        proj_drop: float,
        sr_ratio: int,
    ) -> None:
        super().__init__()
        assert dim % num_heads == 0, f"dim {dim} should be divided by num_heads {num_heads}"
        assert window_size % 2 == 1, "window size must be odd"

        self.num_heads = num_heads
        self.head_dim = dim // num_heads

        self.window_size = window_size
        self.local_len = window_size**2

        pool_h = input_resolution[0] // sr_ratio
        pool_w = input_resolution[1] // sr_ratio
        self.pool_len = pool_h * pool_w

        self.swa_qk_rpb = SWAttention_QK_RPB()
        self.swa_av = SWAttention_AV()
        self.temperature = nn.Parameter(torch.log((torch.ones(num_heads, 1, 1) / 0.24).exp() - 1))

        self.q = nn.Linear(dim, dim, bias=qkv_bias)
        self.query_embedding = nn.Parameter(
            nn.init.trunc_normal_(torch.empty(self.num_heads, 1, self.head_dim), mean=0, std=0.02)
        )
        self.kv = nn.Linear(dim, dim * 2, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        # Components to generate pooled features
        self.pool = nn.AdaptiveAvgPool2d((pool_h, pool_w))
        self.sr = nn.Conv2d(dim, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))
        self.norm = nn.LayerNorm(dim)
        self.act = nn.GELU()

        # MLP to generate continuous relative position bias
        self.cpb_fc1 = nn.Linear(2, 512, bias=True)
        self.cpb_act = nn.ReLU(inplace=True)
        self.cpb_fc2 = nn.Linear(512, num_heads, bias=True)

        # relative bias for local features
        self.relative_pos_bias_local = nn.Parameter(
            nn.init.trunc_normal_(torch.empty(num_heads, self.local_len), mean=0, std=0.0004)
        )

        # Generate padding_mask and sequence length scale
        (local_seq_length, padding_mask) = get_seqlen_and_mask(input_resolution, self.window_size)
        self.seq_length_scale = nn.Buffer(torch.log(local_seq_length + self.pool_len), persistent=False)
        self.padding_mask = nn.Buffer(padding_mask, persistent=False)

        # Dynamic local bias
        self.learnable_tokens = nn.Parameter(
            nn.init.trunc_normal_(torch.empty(num_heads, self.head_dim, self.local_len), mean=0, std=0.02)
        )
        self.learnable_bias = nn.Parameter(torch.zeros(num_heads, 1, self.local_len))

    def forward(
        self, x: torch.Tensor, H: int, W: int, relative_pos_index: torch.Tensor, relative_coords_table: torch.Tensor
    ) -> torch.Tensor:
        (B, N, C) = x.size()

        # Generate queries, normalize them with L2, add query embedding,
        # and then magnify with sequence length scale and temperature.
        # Use softplus function ensuring that the temperature is not lower than 0.
        q_norm = F.normalize(self.q(x).reshape(B, N, self.num_heads, self.head_dim).permute(0, 2, 1, 3), dim=-1)
        q_norm_scaled = (
            (q_norm + self.query_embedding)
            * F.softplus(self.temperature)  # pylint: disable=not-callable
            * self.seq_length_scale
        )

        (attn_local, v_local) = self.swa_qk_rpb(
            self.kv(x),
            q_norm_scaled.contiguous(),
            self.relative_pos_bias_local,
            self.padding_mask,
            self.num_heads,
            self.head_dim,
            self.window_size,
            self.local_len,
            H,
            W,
        )

        # Generate pooled features
        x_ = x.permute(0, 2, 1).reshape(B, -1, H, W).contiguous()
        x_ = self.pool(self.act(self.sr(x_))).reshape(B, -1, self.pool_len).permute(0, 2, 1)
        x_ = self.norm(x_)

        # Generate pooled keys and values
        kv_pool = self.kv(x_).reshape(B, self.pool_len, 2 * self.num_heads, self.head_dim).permute(0, 2, 1, 3)
        (k_pool, v_pool) = kv_pool.chunk(2, dim=1)

        # Use MLP to generate continuous relative positional bias for pooled features.
        pool_bias = (
            self.cpb_fc2(self.cpb_act(self.cpb_fc1(relative_coords_table)))
            .transpose(0, 1)[:, relative_pos_index.view(-1)]
            .view(-1, N, self.pool_len)
        )
        # Compute pooled similarity
        attn_pool = q_norm_scaled @ F.normalize(k_pool, dim=-1).transpose(-2, -1) + pool_bias

        # Concatenate local & pooled similarity matrices and calculate attention weights through the same Softmax
        attn = torch.concat([attn_local, attn_pool], dim=-1).softmax(dim=-1)
        attn = self.attn_drop(attn)

        # Split the attention weights and separately aggregate the values of local & pooled features
        (attn_local, attn_pool) = torch.split(attn, [self.local_len, self.pool_len], dim=-1)

        x_local = self.swa_av(
            q_norm, attn_local, v_local.contiguous(), self.learnable_tokens, self.learnable_bias, self.window_size, H, W
        )

        x_pool = attn_pool @ v_pool
        x = (x_local + x_pool).transpose(1, 2).reshape(B, N, C)

        # Linear projection and output
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class AttentionBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        input_resolution: tuple[int, int],
        window_size: int,
        mlp_ratio: float,
        qkv_bias: bool,
        proj_drop: float,
        attn_drop: float,
        drop_path: float,
        sr_ratio: int,
    ) -> None:
        super().__init__()
        self.norm1 = nn.LayerNorm(dim, eps=1e-6)
        if sr_ratio == 1:
            self.attn = Attention(
                dim, input_resolution, num_heads=num_heads, qkv_bias=qkv_bias, attn_drop=attn_drop, proj_drop=proj_drop
            )
        else:
            self.attn = AggregatedAttention(
                dim,
                input_resolution,
                window_size=window_size,
                num_heads=num_heads,
                qkv_bias=qkv_bias,
                attn_drop=attn_drop,
                proj_drop=proj_drop,
                sr_ratio=sr_ratio,
            )

        self.norm2 = nn.LayerNorm(dim, eps=1e-6)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = ConvolutionalGLU(in_features=dim, hidden_features=mlp_hidden_dim, drop=proj_drop)

        self.drop_path = StochasticDepth(drop_path, mode="row")

    def forward(
        self, x: torch.Tensor, H: int, W: int, relative_pos_index: torch.Tensor, relative_coords_table: torch.Tensor
    ) -> torch.Tensor:
        x = x + self.drop_path(self.attn(self.norm1(x), H, W, relative_pos_index, relative_coords_table))
        x = x + self.drop_path(self.mlp(self.norm2(x), H, W))

        return x


class OverlapPatchEmbed(nn.Module):
    def __init__(self, patch_size: int, stride: int, in_channels: int, embed_dim: int) -> None:
        super().__init__()
        assert patch_size > stride, "Set larger patch_size than stride"

        self.proj = nn.Conv2d(
            in_channels,
            embed_dim,
            kernel_size=(patch_size, patch_size),
            stride=(stride, stride),
            padding=(patch_size // 2, patch_size // 2),
        )
        self.norm = nn.LayerNorm(embed_dim)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor, int, int]:
        x = self.proj(x)
        (_, _, H, W) = x.size()
        x = x.flatten(2).transpose(1, 2)
        x = self.norm(x)

        return (x, H, W)


class TransNeXtStage(nn.Module):
    def __init__(
        self,
        input_resolution: tuple[int, int],
        sr_ratio: int,
        patch_size: int,
        stride: int,
        in_channels: int,
        embed_dim: int,
        num_heads: int,
        window_size: int,
        mlp_ratio: float,
        qkv_bias: bool,
        proj_drop: float,
        attn_drop: float,
        drop_path: list[float],
        depth: int,
    ) -> None:
        super().__init__()

        # Generate relative positional coordinate table and index for each stage
        # to compute continuous relative positional bias
        (relative_pos_index, relative_coords_table) = get_relative_position_cpb(
            query_size=input_resolution, key_size=(input_resolution[0] // sr_ratio, input_resolution[1] // sr_ratio)
        )
        self.relative_pos_index = nn.Buffer(relative_pos_index, persistent=False)
        self.relative_coords_table = nn.Buffer(relative_coords_table, persistent=False)

        self.patch_embed = OverlapPatchEmbed(
            patch_size=patch_size,
            stride=stride,
            in_channels=in_channels,
            embed_dim=embed_dim,
        )

        self.blocks = nn.ModuleList()
        for i in range(depth):
            self.blocks.append(
                AttentionBlock(
                    embed_dim,
                    num_heads=num_heads,
                    input_resolution=input_resolution,
                    window_size=window_size,
                    mlp_ratio=mlp_ratio,
                    qkv_bias=qkv_bias,
                    proj_drop=proj_drop,
                    attn_drop=attn_drop,
                    drop_path=drop_path[i],
                    sr_ratio=sr_ratio,
                )
            )

        self.norm = nn.LayerNorm(embed_dim, eps=1e-6)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        B = x.size(0)
        (x, H, W) = self.patch_embed(x)
        for blk in self.blocks:
            x = blk(x, H, W, self.relative_pos_index, self.relative_coords_table)

        x = self.norm(x)
        x = x.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()

        return x


class TransNeXt(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        patch_size = 4
        sr_ratio = [8, 4, 2, 1]
        mlp_ratios = [8, 8, 4, 4]
        window_size: list[int] = self.config["window_size"]
        depth: list[int] = self.config["depth"]
        embed_dim: list[int] = self.config["embed_dim"]
        num_heads: list[int] = self.config["num_heads"]
        drop_path_rate: float = self.config["drop_path_rate"]

        self.sr_ratio = sr_ratio
        num_stages = len(depth)
        dpr = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(depth)).split(depth)]

        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i in range(num_stages):
            stages[f"stage{i+1}"] = TransNeXtStage(
                input_resolution=(self.size[0] // (2 ** (i + 2)), self.size[1] // (2 ** (i + 2))),
                sr_ratio=sr_ratio[i],
                patch_size=patch_size * 2 - 1 if i == 0 else 3,
                stride=patch_size if i == 0 else 2,
                in_channels=self.input_channels if i == 0 else embed_dim[i - 1],
                embed_dim=embed_dim[i],
                num_heads=num_heads[i],
                window_size=window_size[i],
                mlp_ratio=mlp_ratios[i],
                qkv_bias=True,
                proj_drop=0.0,
                attn_drop=0.0,
                drop_path=dpr[i],
                depth=depth[i],
            )
            return_channels.append(embed_dim[i])

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = embed_dim[-1]
        self.classifier = self.create_classifier()

        # Weights initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, nn.Conv2d):
                fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                fan_out //= m.groups
                nn.init.trunc_normal_(m.weight, std=math.sqrt(2.0 / fan_out))
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, nn.LayerNorm):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        assert dynamic_size is False, "Dynamic size not supported for this network"

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        super().adjust_size(new_size)

        i = 0
        for m in self.body.modules():
            if isinstance(m, TransNeXtStage):
                input_resolution = (new_size[0] // (2 ** (i + 2)), new_size[1] // (2 ** (i + 2)))
                sr_ratio = self.sr_ratio[i]
                (relative_pos_index, relative_coords_table) = get_relative_position_cpb(
                    query_size=input_resolution,
                    key_size=(input_resolution[0] // sr_ratio, input_resolution[1] // sr_ratio),
                )
                m.relative_pos_index = nn.Buffer(relative_pos_index, persistent=False)
                m.relative_coords_table = nn.Buffer(relative_coords_table, persistent=False)

                for blk in m.modules():
                    if isinstance(blk, Attention):
                        blk.seq_length_scale = nn.Buffer(
                            torch.log(torch.tensor(input_resolution[0] * input_resolution[1])), persistent=False
                        )

                    elif isinstance(blk, AggregatedAttention):
                        pool_h = input_resolution[0] // sr_ratio
                        pool_w = input_resolution[1] // sr_ratio
                        blk.pool_len = pool_h * pool_w
                        blk.pool = nn.AdaptiveAvgPool2d((pool_h, pool_w))

                        (local_seq_length, padding_mask) = get_seqlen_and_mask(input_resolution, blk.window_size)
                        blk.seq_length_scale = nn.Buffer(torch.log(local_seq_length + blk.pool_len), persistent=False)
                        blk.padding_mask = nn.Buffer(padding_mask, persistent=False)

                i += 1


registry.register_model_config(
    "transnext_micro",
    TransNeXt,
    config={
        "window_size": [3, 3, 3, 0],
        "depth": [2, 2, 15, 2],
        "embed_dim": [48, 96, 192, 384],
        "num_heads": [2, 4, 8, 16],
        "drop_path_rate": 0.15,
    },
)
registry.register_model_config(
    "transnext_tiny",
    TransNeXt,
    config={
        "window_size": [3, 3, 3, 0],
        "depth": [2, 2, 15, 2],
        "embed_dim": [72, 144, 288, 576],
        "num_heads": [3, 6, 12, 24],
        "drop_path_rate": 0.25,
    },
)
registry.register_model_config(
    "transnext_small",
    TransNeXt,
    config={
        "window_size": [3, 3, 3, 0],
        "depth": [5, 5, 22, 5],
        "embed_dim": [72, 144, 288, 576],
        "num_heads": [3, 6, 12, 24],
        "drop_path_rate": 0.45,
    },
)
registry.register_model_config(
    "transnext_base",
    TransNeXt,
    config={
        "window_size": [3, 3, 3, 0],
        "depth": [5, 5, 23, 5],
        "embed_dim": [96, 192, 384, 768],
        "num_heads": [4, 8, 16, 32],
        "drop_path_rate": 0.6,
    },
)

registry.register_weights(
    "transnext_micro_il-common",
    {
        "description": "TransNeXt micro model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 48.2,
                "sha256": "2945ab2da656e944a44835c9ed04b06339cdc8e5baea4386de1f077a60ec4e0d",
            }
        },
        "net": {"network": "transnext_micro", "tag": "il-common"},
    },
)



"""
Wide ResNet, adapted from
https://github.com/pytorch/vision/blob/main/torchvision/models/resnet.py

Paper "Wide Residual Networks", https://arxiv.org/abs/1605.07146
"""

# Reference license: BSD 3-Clause

from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation

from birder.model_registry import registry
from birder.net.base import DetectorBackbone


class ResidualBlock(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, stride: tuple[int, int], bottle_neck: bool) -> None:
        super().__init__()
        if bottle_neck is True:
            self.block1 = nn.Sequential(
                Conv2dNormActivation(
                    in_channels,
                    out_channels // 2,
                    kernel_size=(1, 1),
                    stride=(1, 1),
                    padding=(0, 0),
                    bias=False,
                ),
                Conv2dNormActivation(
                    out_channels // 2,
                    out_channels // 2,
                    kernel_size=(3, 3),
                    stride=stride,
                    padding=(1, 1),
                    bias=False,
                ),
                nn.Conv2d(
                    out_channels // 2,
                    out_channels,
                    kernel_size=(1, 1),
                    stride=(1, 1),
                    padding=(0, 0),
                    bias=False,
                ),
                nn.BatchNorm2d(out_channels),
            )

        else:
            self.block1 = nn.Sequential(
                Conv2dNormActivation(
                    in_channels, out_channels, kernel_size=(3, 3), stride=stride, padding=(1, 1), bias=False
                ),
                nn.Conv2d(out_channels, out_channels, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False),
                nn.BatchNorm2d(out_channels),
            )

        if in_channels == out_channels:
            self.block2 = nn.Identity()
        else:
            self.block2 = nn.Sequential(
                nn.Conv2d(in_channels, out_channels, kernel_size=(1, 1), stride=stride, padding=(0, 0), bias=False),
                nn.BatchNorm2d(out_channels),
            )

        self.relu = nn.ReLU(inplace=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x
        x = self.block1(x)
        identity = self.block2(identity)
        x += identity
        x = self.relu(x)

        return x


# pylint: disable=invalid-name
class Wide_ResNet(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        bottle_neck = True
        filter_list = [64, 256, 512, 1024, 2048]
        units: list[int] = self.config["units"]

        assert len(units) + 1 == len(filter_list)
        num_unit = len(units)

        self.stem = nn.Sequential(
            Conv2dNormActivation(
                self.input_channels,
                filter_list[0],
                kernel_size=(7, 7),
                stride=(2, 2),
                padding=(3, 3),
                bias=False,
            ),
            nn.MaxPool2d(kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
        )

        # Generate body layers
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i in range(num_unit):
            layers = []
            if i == 0:
                stride = (1, 1)
            else:
                stride = (2, 2)

            layers.append(ResidualBlock(filter_list[i], filter_list[i + 1], stride=stride, bottle_neck=bottle_neck))
            for _ in range(1, units[i]):
                layers.append(
                    ResidualBlock(filter_list[i + 1], filter_list[i + 1], stride=(1, 1), bottle_neck=bottle_neck)
                )

            stages[f"stage{i+1}"] = nn.Sequential(*layers)
            return_channels.append(filter_list[i + 1])

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = filter_list[-1]
        self.classifier = self.create_classifier()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config("wide_resnet_50", Wide_ResNet, config={"units": [3, 4, 6, 3]})
registry.register_model_config("wide_resnet_101", Wide_ResNet, config={"units": [3, 4, 23, 3]})
registry.register_model_config("wide_resnet_152", Wide_ResNet, config={"units": [3, 8, 36, 3]})
registry.register_model_config("wide_resnet_200", Wide_ResNet, config={"units": [3, 24, 36, 3]})
registry.register_model_config("wide_resnet_269", Wide_ResNet, config={"units": [3, 30, 48, 8]})



import logging
import unittest

import torch

from birder import net
from birder import scheduler

logging.disable(logging.CRITICAL)


class TestScheduler(unittest.TestCase):
    def test_cooldown(self) -> None:
        # Just a simple sanity test
        size = (64, 64)
        n = net.MobileNet_v2(3, 10, config={"alpha": 0.5}, size=size)
        optimizer = torch.optim.SGD(n.parameters(), lr=0.1)
        lr_scheduler = scheduler.CooldownLR(optimizer, total_steps=5)

        optimizer.zero_grad()
        optimizer.step()

        lr_scheduler.step()
        last_lr = lr_scheduler.get_last_lr()[0]
        self.assertAlmostEqual(last_lr, 0.08)

        lr_scheduler.step()
        last_lr = lr_scheduler.get_last_lr()[0]
        self.assertAlmostEqual(last_lr, 0.06)

        lr_scheduler.step()
        last_lr = lr_scheduler.get_last_lr()[0]
        self.assertAlmostEqual(last_lr, 0.04)

        lr_scheduler.step()
        last_lr = lr_scheduler.get_last_lr()[0]
        self.assertAlmostEqual(last_lr, 0.02)

        lr_scheduler.step()
        last_lr = lr_scheduler.get_last_lr()[0]
        self.assertAlmostEqual(last_lr, lr_scheduler.eps)

        lr_scheduler.step()
        last_lr = lr_scheduler.get_last_lr()[0]
        self.assertEqual(last_lr, 0.0)



"""
Sequencer, adapted from
https://github.com/okojoalg/sequencer/blob/main/models/two_dim_sequencer.py
and
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/sequencer.py

Paper "Sequencer: Deep LSTM for Image Classification", https://arxiv.org/abs/2205.01972

Changes from original:
* Only supports the "cat" union (simplify the implementation)
"""

# Reference license: Apache-2.0 (both)

from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import MLP
from torchvision.ops import Permute
from torchvision.ops import StochasticDepth

from birder.model_registry import registry
from birder.net.base import BaseNet


class LSTM2d(nn.Module):
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_layers: int,
    ) -> None:
        super().__init__()

        self.input_size = input_size
        self.hidden_size = hidden_size
        self.output_size = 2 * hidden_size

        self.fc = nn.Linear(2 * self.output_size, input_size)
        self.rnn_v = nn.LSTM(
            input_size,
            hidden_size,
            num_layers,
            batch_first=True,
            bias=True,
            bidirectional=True,
        )
        self.rnn_h = nn.LSTM(
            input_size,
            hidden_size,
            num_layers,
            batch_first=True,
            bias=True,
            bidirectional=True,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, H, W, C) = x.shape

        v = x.permute(0, 2, 1, 3)
        v = v.reshape(-1, H, C)
        (v, _) = self.rnn_v(v)
        v = v.reshape(B, W, H, -1)
        v = v.permute(0, 2, 1, 3)

        h = x.reshape(-1, W, C)
        (h, _) = self.rnn_h(h)
        h = h.reshape(B, H, W, -1)

        x = torch.concat([v, h], dim=-1)
        x = self.fc(x)

        return x


class Sequencer2dBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        hidden_size: int,
        mlp_ratio: float,
        num_layers: int,
        drop: float,
        drop_path: float,
    ) -> None:
        super().__init__()
        self.norm1 = nn.LayerNorm(dim, eps=1e-6)
        self.rnn = LSTM2d(dim, hidden_size, num_layers=num_layers)
        self.drop_path = StochasticDepth(drop_path, mode="row")
        self.norm2 = nn.LayerNorm(dim, eps=1e-6)
        self.mlp = MLP(dim, [int(mlp_ratio * dim), dim], activation_layer=nn.GELU, dropout=drop)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.drop_path(self.rnn(self.norm1(x)))
        x = x + self.drop_path(self.mlp(self.norm2(x)))

        return x


class Downsample2d(nn.Module):
    def __init__(self, input_dim: int, output_dim: int, patch_size: int) -> None:
        super().__init__()
        self.down = nn.Conv2d(
            input_dim, output_dim, kernel_size=(patch_size, patch_size), stride=(patch_size, patch_size), padding=(0, 0)
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x.permute(0, 3, 1, 2)
        x = self.down(x)
        x = x.permute(0, 2, 3, 1)

        return x


class Sequencer2dStage(nn.Module):
    def __init__(
        self,
        dim: int,
        dim_out: int,
        depth: int,
        patch_size: int,
        hidden_size: int,
        mlp_ratio: float,
        downsample: bool,
        num_layers: int,
        drop: float,
        drop_path: float,
    ) -> None:
        super().__init__()
        if downsample is True:
            self.downsample = Downsample2d(dim, dim_out, patch_size)
        else:
            assert dim == dim_out
            self.downsample = nn.Identity()

        blocks = []
        for _ in range(depth):
            blocks.append(
                Sequencer2dBlock(
                    dim_out,
                    hidden_size,
                    mlp_ratio=mlp_ratio,
                    num_layers=num_layers,
                    drop=drop,
                    drop_path=drop_path,
                )
            )

        self.blocks = nn.Sequential(*blocks)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.downsample(x)
        x = self.blocks(x)

        return x


class Sequencer2d(BaseNet):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        hidden_sizes = [48, 96, 96, 96]
        mlp_ratios = [3.0, 3.0, 3.0, 3.0]
        num_rnn_layers = 1
        drop_rate = 0.0
        layers: list[int] = self.config["layers"]
        patch_sizes: list[int] = self.config["patch_sizes"]
        embed_dims: list[int] = self.config["embed_dims"]
        drop_path_rate: float = self.config["drop_path_rate"]

        assert len(layers) == len(patch_sizes) == len(embed_dims) == len(hidden_sizes) == len(mlp_ratios)

        self.stem = nn.Sequential(
            nn.Conv2d(
                self.input_channels,
                embed_dims[0],
                kernel_size=(patch_sizes[0], patch_sizes[0]),
                stride=(patch_sizes[0], patch_sizes[0]),
                padding=(0, 0),
                bias=True,
            ),
            Permute([0, 2, 3, 1]),
        )

        stages = []
        prev_dim = embed_dims[0]
        for idx, embed_dim in enumerate(embed_dims):
            sd_prob = drop_path_rate * float(idx) / len(layers)

            stages += [
                Sequencer2dStage(
                    prev_dim,
                    embed_dim,
                    depth=layers[idx],
                    patch_size=patch_sizes[idx],
                    hidden_size=hidden_sizes[idx],
                    mlp_ratio=mlp_ratios[idx],
                    downsample=idx > 0,
                    num_layers=num_rnn_layers,
                    drop=drop_rate,
                    drop_path=sd_prob,
                )
            ]
            prev_dim = embed_dim

        self.body = nn.Sequential(*stages)
        self.features = nn.Sequential(
            nn.LayerNorm(prev_dim, eps=1e-6),
            Permute([0, 3, 1, 2]),  # B H W C -> B C H W)
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.embedding_size = prev_dim
        self.classifier = self.create_classifier()

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config(
    "sequencer2d_s",
    Sequencer2d,
    config={
        "layers": [4, 3, 8, 3],
        "patch_sizes": [7, 2, 1, 1],
        "embed_dims": [192, 384, 384, 384],
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "sequencer2d_m",
    Sequencer2d,
    config={
        "layers": [4, 3, 14, 3],
        "patch_sizes": [7, 2, 1, 1],
        "embed_dims": [192, 384, 384, 384],
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "sequencer2d_l",
    Sequencer2d,
    config={
        "layers": [8, 8, 16, 4],
        "patch_sizes": [7, 2, 1, 1],
        "embed_dims": [192, 384, 384, 384],
        "drop_path_rate": 0.0,
    },
)



"""
RegionViT, adapted from
https://github.com/IBM/RegionViT/blob/master/regionvit/regionvit.py

Paper "RegionViT: Regional-to-Local Attention for Vision Transformers", https://arxiv.org/abs/2106.02689

Changes from original:
* Default size is 256 instead of 224 (window size is 8 instead of 7),
  note that the window does NOT change with resolution changes
"""

# Reference license: Apache-2.0

from collections import OrderedDict
from typing import Any
from typing import Literal
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import MLP
from torchvision.ops import StochasticDepth

from birder.model_registry import registry
from birder.net.base import DetectorBackbone


def convert_to_flatten_layout(
    cls_tokens: torch.Tensor, patch_tokens: torch.Tensor, ws: int
) -> tuple[torch.Tensor, Optional[torch.Tensor], int, int, int, int, int, int]:
    # Padding if added will be at the bottom right
    (B, C, H, W) = patch_tokens.size()
    (_, _, h_ks, w_ks) = cls_tokens.size()
    need_mask = False
    p_l = 0
    p_r = 0
    p_t = 0
    p_b = 0
    if H % (h_ks * ws) != 0 or W % (w_ks * ws) != 0:
        p_r = w_ks * ws - W
        p_b = h_ks * ws - H
        patch_tokens = F.pad(patch_tokens, (p_l, p_r, p_t, p_b))
        need_mask = True

    (B, C, H, W) = patch_tokens.size()
    kernel_size = (H // h_ks, W // w_ks)
    tmp = F.unfold(patch_tokens, kernel_size=kernel_size, dilation=(1, 1), padding=(0, 0), stride=kernel_size)
    patch_tokens = tmp.transpose(1, 2).reshape(-1, C, kernel_size[0] * kernel_size[1]).transpose(-2, -1)

    if need_mask is True:
        (bh_sk_s, ksks, C) = patch_tokens.size()
        h_s = H // ws
        w_s = W // ws
        mask = torch.ones(bh_sk_s // B, 1 + ksks, 1 + ksks, device=patch_tokens.device, dtype=torch.float)
        right = torch.zeros(1 + ksks, 1 + ksks, device=patch_tokens.device, dtype=torch.float)
        tmp = torch.zeros(ws, ws, device=patch_tokens.device, dtype=torch.float)
        tmp[0 : (ws - p_r), 0 : (ws - p_r)] = 1.0
        tmp = tmp.repeat(ws, ws)
        right[1:, 1:] = tmp
        right[0, 0] = 1
        right[0, 1:] = torch.tensor([1.0] * (ws - p_r) + [0.0] * p_r).repeat(ws).to(right.device)
        right[1:, 0] = torch.tensor([1.0] * (ws - p_r) + [0.0] * p_r).repeat(ws).to(right.device)
        bottom = torch.zeros_like(right)
        bottom[0 : ws * (ws - p_b) + 1, 0 : ws * (ws - p_b) + 1] = 1.0
        bottom_right = right.clone()
        bottom_right[0 : ws * (ws - p_b) + 1, 0 : ws * (ws - p_b) + 1] = 1.0

        mask[w_s - 1 : (h_s - 1) * w_s : w_s, ...] = right
        mask[(h_s - 1) * w_s :, ...] = bottom
        mask[-1, ...] = bottom_right
        mask = mask.repeat(B, 1, 1)
    else:
        mask = None

    cls_tokens = cls_tokens.flatten(2).transpose(-2, -1)  # (N)x(H/sxK/s)xC
    cls_tokens = cls_tokens.reshape(-1, 1, cls_tokens.size(-1))  # (NxH/sxK/s)x1xC

    out = torch.concat((cls_tokens, patch_tokens), dim=1)

    return (out, mask, p_r, p_b, B, C, H, W)


def convert_to_spatial_layout(
    out: torch.Tensor,
    B: int,
    C: int,
    H: int,
    W: int,
    kernel_size: tuple[int, int],
    mask: Optional[torch.Tensor],
    p_r: int,
    p_b: int,
) -> tuple[torch.Tensor, torch.Tensor]:
    cls_tokens = out[:, 0:1, ...]
    patch_tokens = out[:, 1:, ...]

    h_ks = H // kernel_size[0]
    w_ks = W // kernel_size[1]

    cls_tokens = cls_tokens.reshape(B, -1, C).transpose(-2, -1).reshape(B, C, h_ks, w_ks)
    patch_tokens = patch_tokens.transpose(1, 2).reshape((B, -1, kernel_size[0] * kernel_size[1] * C)).transpose(1, 2)
    patch_tokens = F.fold(patch_tokens, (H, W), kernel_size=kernel_size, stride=kernel_size, padding=(0, 0))

    if mask is not None:
        if p_b > 0:
            patch_tokens = patch_tokens[:, :, :-p_b, :]
        if p_r > 0:
            patch_tokens = patch_tokens[:, :, :, :-p_r]

    return (cls_tokens, patch_tokens)


class SequentialWithTwo(nn.Sequential):
    def forward(  # pylint: disable=arguments-differ
        self, cls_tokens: torch.Tensor, patch_tokens: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor]:
        for module in self:
            (cls_tokens, patch_tokens) = module(cls_tokens, patch_tokens)

        return (cls_tokens, patch_tokens)


class LayerNorm2d(nn.Module):
    def __init__(self, channels: int, eps: float = 1e-5) -> None:
        super().__init__()

        self.eps = nn.Buffer(torch.tensor(eps))
        self.weight = nn.Parameter(torch.ones(1, channels, 1, 1))
        self.bias = nn.Parameter(torch.zeros(1, channels, 1, 1))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        mean = x.mean(1, keepdim=True)
        std = torch.sqrt(x.var(1, unbiased=False, keepdim=True) + self.eps)
        out = (x - mean) / std
        out = out * self.weight + self.bias

        return out


class AttentionWithRelPos(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        qkv_bias: bool,
        attn_drop: float,
        proj_drop: float,
        attn_map_dim: tuple[int, int],
        num_cls_tokens: int,
    ) -> None:
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        self.num_cls_tokens = num_cls_tokens
        one_dim = attn_map_dim[0]
        rel_pos_dim = 2 * one_dim - 1
        self.rel_pos = nn.Parameter(torch.zeros(num_heads, rel_pos_dim**2))
        tmp = torch.arange(rel_pos_dim**2).reshape((rel_pos_dim, rel_pos_dim))
        out = []
        offset_x = offset_y = one_dim // 2
        for y in range(one_dim):
            for x in range(one_dim):
                for dy in range(one_dim):
                    for dx in range(one_dim):
                        out.append(tmp[dy - y + offset_y, dx - x + offset_x])

        self.rel_pos_index = nn.Buffer(torch.tensor(out, dtype=torch.long))

        # Weight initialization
        nn.init.trunc_normal_(self.rel_pos, std=0.02)

    def forward(self, x: torch.Tensor, patch_attn: bool = False, mask: Optional[torch.Tensor] = None) -> torch.Tensor:
        (B, N, C) = x.size()
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        (q, k, v) = qkv.unbind(0)

        attn = (q @ k.transpose(-2, -1)) * self.scale

        if patch_attn is True:
            rel_pos = self.rel_pos[:, self.rel_pos_index].reshape(
                self.num_heads, N - self.num_cls_tokens, N - self.num_cls_tokens
            )
            attn[:, :, self.num_cls_tokens :, self.num_cls_tokens :] = (
                attn[:, :, self.num_cls_tokens :, self.num_cls_tokens :] + rel_pos
            )

        if mask is not None:
            mask = mask.unsqueeze(1).expand(-1, self.num_heads, -1, -1)
            attn = attn.masked_fill(mask == 0, -65000)

        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class PatchEmbed(nn.Module):
    def __init__(
        self,
        patch_size: tuple[int, int],
        in_channels: int,
        embed_dim: int,
        patch_conv_type: Literal["3conv", "1conv", "linear"],
    ) -> None:
        super().__init__()
        self.patch_size = patch_size

        if patch_conv_type == "3conv":
            assert patch_size[0] == 4 and patch_size[1] == 4
            self.proj = nn.Sequential(
                nn.Conv2d(in_channels, embed_dim // 4, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
                LayerNorm2d(embed_dim // 4),
                nn.GELU(),
                nn.Conv2d(embed_dim // 4, embed_dim // 2, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
                LayerNorm2d(embed_dim // 2),
                nn.GELU(),
                nn.Conv2d(embed_dim // 2, embed_dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1)),
            )
        elif patch_conv_type == "1conv":
            kernel_size = (2 * patch_size[0], 2 * patch_size[1])
            stride = (patch_size[0], patch_size[1])
            padding = (patch_size[0] - 1, patch_size[1] - 1)
            self.proj = nn.Conv2d(in_channels, embed_dim, kernel_size=kernel_size, stride=stride, padding=padding)
        elif patch_conv_type == "linear":
            kernel_size = patch_size
            stride = patch_size
            padding = (0, 0)
            self.proj = nn.Conv2d(in_channels, embed_dim, kernel_size=kernel_size, stride=stride, padding=padding)
        else:
            raise ValueError("Unknown patch_conv_type")

    def forward(self, x: torch.Tensor, extra_padding: bool = False) -> torch.Tensor:
        (_, _, H, W) = x.size()
        if extra_padding and (H % self.patch_size[0] != 0 or W % self.patch_size[1] != 0):
            p_l = (self.patch_size[1] - W % self.patch_size[1]) // 2
            p_r = (self.patch_size[1] - W % self.patch_size[1]) - p_l
            p_t = (self.patch_size[0] - H % self.patch_size[0]) // 2
            p_b = (self.patch_size[0] - H % self.patch_size[0]) - p_t
            x = F.pad(x, (p_l, p_r, p_t, p_b))

        return self.proj(x)


class R2LAttentionFFN(nn.Module):
    def __init__(
        self,
        input_channels: int,
        output_channels: int,
        window_size: tuple[int, int],
        num_heads: int,
        mlp_ratio: float,
        qkv_bias: bool,
        drop_path: float,
        attn_drop: float,
        drop: float,
    ) -> None:
        super().__init__()

        self.norm0 = nn.LayerNorm(input_channels, eps=1e-6)
        self.norm1 = nn.LayerNorm(input_channels, eps=1e-6)
        self.attn = AttentionWithRelPos(
            input_channels,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            attn_drop=attn_drop,
            proj_drop=drop,
            attn_map_dim=window_size,
            num_cls_tokens=1,
        )

        self.drop_path = StochasticDepth(drop_path, mode="row")
        self.norm2 = nn.LayerNorm(input_channels, eps=1e-6)
        self.mlp = MLP(
            input_channels, [int(input_channels * mlp_ratio), output_channels], activation_layer=nn.GELU, dropout=drop
        )

        if input_channels != output_channels:
            self.expand = nn.Sequential(
                nn.LayerNorm(input_channels, eps=1e-6),
                nn.GELU(),
                nn.Linear(input_channels, output_channels),
            )
        else:
            self.expand = nn.Identity()

    def forward(self, x: torch.Tensor, mask: Optional[torch.Tensor], batch_size: int) -> torch.Tensor:
        cls_tokens = x[:, 0:1, ...]
        C = cls_tokens.size(-1)
        cls_tokens = cls_tokens.reshape(batch_size, -1, C)  # (N)x(H/sxW/s)xC
        cls_tokens = cls_tokens + self.drop_path(self.attn(self.norm0(cls_tokens)))  # (N)x(H/sxK/s)xC

        cls_tokens = cls_tokens.reshape(-1, 1, C)  # (NxH/sxK/s)x1xC

        out = torch.concat((cls_tokens, x[:, 1:, ...]), dim=1)

        out = out + self.drop_path(self.attn(self.norm1(out), patch_attn=True, mask=mask))
        identity = self.expand(out)
        out = identity + self.drop_path(self.mlp(self.norm2(out)))

        return out


class Projection(nn.Module):
    def __init__(self, input_channels: int, output_channels: int, pool: bool) -> None:
        super().__init__()
        if pool is True:
            kernel_size = 3
            stride = 2
            padding = 1
        else:
            kernel_size = 1
            stride = 1
            padding = 0

        if input_channels == output_channels and pool is False:
            self.proj = nn.Identity()
        else:
            self.proj = nn.Sequential(
                LayerNorm2d(input_channels),
                nn.GELU(),
                nn.Conv2d(
                    input_channels,
                    output_channels,
                    kernel_size=(kernel_size, kernel_size),
                    stride=(stride, stride),
                    padding=(padding, padding),
                    groups=input_channels,
                ),
            )

    def forward(self, cls_tokens: torch.Tensor, patch_tokens: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        cls_tokens = self.proj(cls_tokens)
        patch_tokens = self.proj(patch_tokens)
        return (cls_tokens, patch_tokens)


class ConvAttStage(nn.Module):
    def __init__(
        self,
        input_channels: int,
        output_channels: int,
        window_size: tuple[int, int],
        num_blocks: int,
        num_heads: int,
        mlp_ratio: float,
        qkv_bias: bool,
        pool: bool,
        drop_path_rate: list[float],
        attn_drop_rate: float,
        drop_rate: float,
    ):
        super().__init__()
        assert window_size[0] == window_size[1]
        self.proj = Projection(input_channels, output_channels, pool=pool)

        self.blocks = nn.ModuleList()
        for i in range(num_blocks):
            self.blocks.append(
                R2LAttentionFFN(
                    output_channels,
                    output_channels,
                    window_size,
                    num_heads,
                    mlp_ratio,
                    qkv_bias,
                    drop_path=drop_path_rate[i],
                    attn_drop=attn_drop_rate,
                    drop=drop_rate,
                )
            )

        self.ws = window_size

    def forward(self, cls_tokens: torch.Tensor, patch_tokens: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        (cls_tokens, patch_tokens) = self.proj(cls_tokens, patch_tokens)
        (out, mask, p_r, p_b, B, C, H, W) = convert_to_flatten_layout(cls_tokens, patch_tokens, self.ws[0])
        for blk in self.blocks:
            out = blk(out, mask, B)

        (cls_tokens, patch_tokens) = convert_to_spatial_layout(out, B, C, H, W, self.ws, mask, p_r, p_b)

        return (cls_tokens, patch_tokens)


class RegionViT(DetectorBackbone):
    default_size = (256, 256)

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        mlp_ratio = 4.0
        patch_size = (4, 4)
        patch_conv_type: Literal["3conv", "1conv", "linear"] = self.config["patch_conv_type"]
        depths: list[int] = self.config["depths"]
        embed_dims: list[int] = self.config["embed_dims"]
        num_heads: list[int] = self.config["num_heads"]
        window_size: int = self.config["window_size"]
        drop_path_rate: float = self.config["drop_path_rate"]

        self.patch_embed = PatchEmbed(patch_size, self.input_channels, embed_dims[0], patch_conv_type)
        self.cls_token = PatchEmbed(
            (patch_size[0] * window_size, patch_size[1] * window_size),
            self.input_channels,
            embed_dims[0],
            "linear",
        )

        num_stages = len(depths)
        dpr = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(depths)).split(depths)]
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i in range(num_stages):
            stages[f"stage{i+1}"] = ConvAttStage(
                embed_dims[i],
                embed_dims[i + 1],
                window_size=(window_size, window_size),
                num_blocks=depths[i],
                num_heads=num_heads[i],
                mlp_ratio=mlp_ratio,
                qkv_bias=True,
                pool=i > 0,
                drop_path_rate=dpr[i],
                attn_drop_rate=0.0,
                drop_rate=0.0,
            )
            return_channels.append(embed_dims[i + 1])

        self.body = SequentialWithTwo(stages)
        self.norm = nn.LayerNorm(embed_dims[-1])
        self.return_channels = return_channels
        self.embedding_size = embed_dims[-1]
        self.classifier = self.create_classifier()

        # Weights initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, nn.LayerNorm):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

    def freeze(self, freeze_classifier: bool = True, unfreeze_features: bool = False) -> None:
        for param in self.parameters():
            param.requires_grad = False

        if freeze_classifier is False:
            for param in self.classifier.parameters():
                param.requires_grad = True
        if unfreeze_features is True:
            for param in self.norm.parameters():
                param.requires_grad = True

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        o_x = x
        x = self.patch_embed(x)
        cls_tokens = self.cls_token(o_x, extra_padding=True)

        out = {}
        for name, module in self.body.named_children():
            (cls_tokens, x) = module(cls_tokens, x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.patch_embed.parameters():
            param.requires_grad = False
        for param in self.cls_token.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        o_x = x
        x = self.patch_embed(x)
        cls_tokens = self.cls_token(o_x, extra_padding=True)
        (cls_tokens, x) = self.body(cls_tokens, x)

        return (cls_tokens, x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        (cls_tokens, _) = self.forward_features(x)

        (N, C, _, _) = cls_tokens.size()
        cls_tokens = cls_tokens.reshape(N, C, -1).transpose(1, 2)
        cls_tokens = self.norm(cls_tokens)
        out = torch.mean(cls_tokens, dim=1)

        return out


registry.register_model_config(
    "regionvit_t",
    RegionViT,
    config={
        "patch_conv_type": "3conv",
        "depths": [2, 2, 8, 2],
        "embed_dims": [64, 64, 128, 256, 512],
        "num_heads": [2, 4, 8, 16],
        "window_size": 8,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "regionvit_s",
    RegionViT,
    config={
        "patch_conv_type": "3conv",
        "depths": [2, 2, 8, 2],
        "embed_dims": [96, 96, 192, 384, 768],
        "num_heads": [3, 6, 12, 24],
        "window_size": 8,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "regionvit_m",
    RegionViT,
    config={
        "patch_conv_type": "1conv",
        "depths": [2, 2, 14, 2],
        "embed_dims": [96, 96, 192, 384, 768],
        "num_heads": [3, 6, 12, 24],
        "window_size": 8,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "regionvit_b",
    RegionViT,
    config={
        "patch_conv_type": "1conv",
        "depths": [2, 2, 14, 2],
        "embed_dims": [128, 128, 256, 512, 1024],
        "num_heads": [4, 8, 16, 32],
        "window_size": 8,
        "drop_path_rate": 0.1,
    },
)

registry.register_weights(
    "regionvit_t_il-common",
    {
        "description": "RegionViT tiny model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 52.3,
                "sha256": "06c9aff4086d5cb892da6dc20d7cf383e1c7e0a112948ec235615b1a4930fa5f",
            }
        },
        "net": {"network": "regionvit_t", "tag": "il-common"},
    },
)



"""
MobileNet v1, adapted from
https://github.com/apache/mxnet/blob/1.9.1/example/image-classification/symbols/mobilenet.py

Paper "MobileNets: Efficient Convolutional Neural Networks for Mobile Vision Applications",
https://arxiv.org/abs/1704.04861
"""

# Reference license: Apache-2.0

from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation

from birder.model_registry import registry
from birder.net.base import DetectorBackbone


class DepthwiseSeparableNormConv2d(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, stride: tuple[int, int]) -> None:
        super().__init__()
        self.dpw_bn_conv = nn.Sequential(
            Conv2dNormActivation(
                in_channels,
                in_channels,
                kernel_size=(3, 3),
                stride=stride,
                padding=(1, 1),
                groups=in_channels,
                bias=False,
            ),
            Conv2dNormActivation(
                in_channels,
                out_channels,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
            ),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.dpw_bn_conv(x)


# pylint: disable=invalid-name
class MobileNet_v1(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        alpha: float = self.config["alpha"]

        base = int(32 * alpha)
        self.stem = nn.Sequential(
            Conv2dNormActivation(
                self.input_channels, base, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), bias=False
            ),
            Conv2dNormActivation(
                base, base, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=base, bias=False
            ),
            Conv2dNormActivation(base, base * 2, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False),
        )

        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []

        stages["stage1"] = nn.Sequential(
            DepthwiseSeparableNormConv2d(base * 2, base * 4, stride=(2, 2)),
            DepthwiseSeparableNormConv2d(base * 4, base * 4, stride=(1, 1)),
        )
        return_channels.append(base * 4)

        stages["stage2"] = nn.Sequential(
            DepthwiseSeparableNormConv2d(base * 4, base * 8, stride=(2, 2)),
            DepthwiseSeparableNormConv2d(base * 8, base * 8, stride=(1, 1)),
        )
        return_channels.append(base * 8)

        stages["stage3"] = nn.Sequential(
            DepthwiseSeparableNormConv2d(base * 8, base * 16, stride=(2, 2)),
            DepthwiseSeparableNormConv2d(base * 16, base * 16, stride=(1, 1)),
            DepthwiseSeparableNormConv2d(base * 16, base * 16, stride=(1, 1)),
            DepthwiseSeparableNormConv2d(base * 16, base * 16, stride=(1, 1)),
            DepthwiseSeparableNormConv2d(base * 16, base * 16, stride=(1, 1)),
            DepthwiseSeparableNormConv2d(base * 16, base * 16, stride=(1, 1)),
        )
        return_channels.append(base * 16)

        stages["stage4"] = nn.Sequential(
            DepthwiseSeparableNormConv2d(base * 16, base * 32, stride=(2, 2)),
            # 32
            DepthwiseSeparableNormConv2d(base * 32, base * 32, stride=(1, 1)),
        )
        return_channels.append(base * 32)

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            Conv2dNormActivation(
                base * 32,
                base * 32,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
                activation_layer=None,
            ),
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = base * 32
        self.classifier = self.create_classifier()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config("mobilenet_v1_0_25", MobileNet_v1, config={"alpha": 0.25})
registry.register_model_config("mobilenet_v1_0_5", MobileNet_v1, config={"alpha": 0.5})
registry.register_model_config("mobilenet_v1_0_75", MobileNet_v1, config={"alpha": 0.75})
registry.register_model_config("mobilenet_v1_1_0", MobileNet_v1, config={"alpha": 1.0})



"""
MobileViT v2, adapted from
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/mobilevit.py
and
https://github.com/apple/ml-cvnets/blob/main/cvnets/models/classification/mobilevit_v2.py

Paper "Separable Self-attention for Mobile Vision Transformers",
https://arxiv.org/abs/2206.02680
"""

# Reference license: Apache-2.0 and Apple open source (see license at reference)

import math
from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import StochasticDepth

from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.mobilenet_v2 import InvertedResidual


class LinearSelfAttention(nn.Module):
    def __init__(
        self,
        embed_dim: int,
        attn_drop: float,
        proj_drop: float,
        bias: bool,
    ) -> None:
        super().__init__()
        self.embed_dim = embed_dim

        self.qkv_proj = nn.Conv2d(
            in_channels=embed_dim,
            out_channels=1 + (2 * embed_dim),
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            bias=bias,
        )
        self.attn_drop = nn.Dropout(attn_drop)
        self.out_proj = nn.Conv2d(
            in_channels=embed_dim,
            out_channels=embed_dim,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            bias=bias,
        )
        self.out_drop = nn.Dropout(proj_drop)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # [B, C, P, N] --> [B, h + 2d, P, N]
        qkv = self.qkv_proj(x)

        # Project x into query, key and value
        # Query --> [B, 1, P, N]
        # value, key --> [B, d, P, N]
        (query, key, value) = qkv.split([1, self.embed_dim, self.embed_dim], dim=1)

        # apply softmax along N dimension
        context_scores = F.softmax(query, dim=-1)
        context_scores = self.attn_drop(context_scores)

        # Compute context vector
        # [B, d, P, N] x [B, 1, P, N] -> [B, d, P, N] --> [B, d, P, 1]
        context_vector = (key * context_scores).sum(dim=-1, keepdim=True)

        # combine context vector with values
        # [B, d, P, N] * [B, d, P, 1] --> [B, d, P, N]
        out = F.relu(value) * context_vector.expand_as(value)
        out = self.out_proj(out)
        out = self.out_drop(out)

        return out


class LinearTransformerBlock(nn.Module):
    def __init__(
        self,
        embed_dim: int,
        mlp_ratio: float,
        drop: float,
        attn_drop: float,
        drop_path: float,
    ) -> None:
        super().__init__()
        self.norm1 = nn.GroupNorm(num_groups=1, num_channels=embed_dim)
        self.attn = LinearSelfAttention(embed_dim=embed_dim, attn_drop=attn_drop, proj_drop=drop, bias=True)
        self.drop_path1 = StochasticDepth(drop_path, mode="row")

        self.norm2 = nn.GroupNorm(num_groups=1, num_channels=embed_dim)
        self.mlp = nn.Sequential(
            nn.Conv2d(
                embed_dim, int(embed_dim * mlp_ratio), kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=True
            ),
            nn.SiLU(),
            nn.Dropout(drop),
            nn.Conv2d(
                int(embed_dim * mlp_ratio), embed_dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=True
            ),
        )
        self.drop_path2 = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.drop_path1(self.attn(self.norm1(x)))
        x = x + self.drop_path2(self.mlp(self.norm2(x)))

        return x


class MobileVitBlock(nn.Module):
    def __init__(
        self,
        channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        transformer_dim: int,
        transformer_depth: int,
        mlp_ratio: float,
        patch_size: tuple[int, int],
        attn_drop: float,
        drop: float = 0.0,
        drop_path_rate: float = 0.0,
    ) -> None:
        super().__init__()

        self.conv_kxk = Conv2dNormActivation(
            channels,
            channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=((kernel_size[0] - 1) // 2, (kernel_size[1] - 1) // 2),
            groups=channels,
            activation_layer=nn.SiLU,
        )
        self.conv_1x1 = nn.Conv2d(
            channels, transformer_dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False
        )
        self.transformer = nn.Sequential(
            *[
                LinearTransformerBlock(
                    embed_dim=transformer_dim,
                    mlp_ratio=mlp_ratio,
                    drop=drop,
                    attn_drop=attn_drop,
                    drop_path=drop_path_rate,
                )
                for _ in range(transformer_depth)
            ]
        )
        self.norm = nn.GroupNorm(num_groups=1, num_channels=transformer_dim)

        self.conv_proj = Conv2dNormActivation(
            transformer_dim,
            channels,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            activation_layer=nn.SiLU,
        )

        self.patch_size = patch_size
        self.patch_area = self.patch_size[0] * self.patch_size[1]

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, C, H, W) = x.shape
        (patch_h, patch_w) = self.patch_size
        new_h = math.ceil(H / patch_h) * patch_h
        new_w = math.ceil(W / patch_w) * patch_w
        num_patch_h, num_patch_w = new_h // patch_h, new_w // patch_w  # n_h, n_w
        num_patches = num_patch_h * num_patch_w  # N
        if new_h != H or new_w != W:
            x = F.interpolate(x, size=(new_h, new_w), mode="bilinear", align_corners=True)

        # Local representation
        x = self.conv_kxk(x)
        x = self.conv_1x1(x)

        # Unfold (feature map -> patches), [B, C, H, W] -> [B, C, P, N]
        C = x.shape[1]
        x = x.reshape(B, C, num_patch_h, patch_h, num_patch_w, patch_w).permute(0, 1, 3, 5, 2, 4)
        x = x.reshape(B, C, -1, num_patches)

        # Global representations
        x = self.transformer(x)
        x = self.norm(x)

        # Fold (patches -> feature map), [B, C, P, N] --> [B, C, H, W]
        x = x.reshape(B, C, patch_h, patch_w, num_patch_h, num_patch_w).permute(0, 1, 4, 2, 5, 3)
        x = x.reshape(B, C, num_patch_h * patch_h, num_patch_w * patch_w)

        x = self.conv_proj(x)

        return x


# pylint: disable=invalid-name
class MobileViT_v2(DetectorBackbone):
    default_size = (256, 256)

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        width_factor: float = self.config["width_factor"]

        patch_size = (2, 2)
        attn_drop = 0.1
        depths = [2, 4, 3]
        expansion = 2
        stem_channels = int(32 * width_factor)
        channels: tuple[int, ...] = (64, 128, 256, 384, 512)
        channels = tuple([int(c * width_factor) for c in channels])  # pylint: disable=consider-using-generator
        dims: tuple[int, ...] = (128, 192, 256)
        dims = tuple([int(d * width_factor) for d in dims])  # pylint: disable=consider-using-generator

        self.stem = Conv2dNormActivation(
            self.input_channels,
            stem_channels,
            kernel_size=(3, 3),
            stride=(2, 2),
            padding=(1, 1),
            activation_layer=nn.SiLU,
            bias=True,
        )

        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []

        stages["stage1"] = nn.Sequential(
            InvertedResidual(
                stem_channels,
                channels[0],
                kernel_size=(3, 3),
                stride=(1, 1),
                padding=(1, 1),
                expansion_factor=expansion,
                shortcut=False,
                activation_layer=nn.SiLU,
            ),
            InvertedResidual(
                channels[0],
                channels[1],
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=(1, 1),
                expansion_factor=expansion,
                shortcut=False,
                activation_layer=nn.SiLU,
            ),
            InvertedResidual(
                channels[1],
                channels[1],
                kernel_size=(3, 3),
                stride=(1, 1),
                padding=(1, 1),
                expansion_factor=expansion,
                shortcut=True,
                activation_layer=nn.SiLU,
            ),
        )
        return_channels.append(channels[1])

        for idx in range(2, len(channels)):
            stages[f"stage{idx}"] = nn.Sequential(
                InvertedResidual(
                    channels[idx - 1],
                    channels[idx],
                    kernel_size=(3, 3),
                    stride=(2, 2),
                    padding=(1, 1),
                    expansion_factor=expansion,
                    shortcut=False,
                    activation_layer=nn.SiLU,
                ),
                MobileVitBlock(
                    channels[idx],
                    kernel_size=(3, 3),
                    stride=(1, 1),
                    transformer_dim=dims[idx - 2],
                    transformer_depth=depths[idx - 2],
                    mlp_ratio=2,
                    patch_size=patch_size,
                    attn_drop=attn_drop,
                ),
            )
            return_channels.append(channels[idx])

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = channels[-1]
        self.classifier = self.create_classifier()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def create_classifier(self, embed_dim: Optional[int] = None) -> nn.Module:
        if self.num_classes == 0:
            return nn.Identity()

        if embed_dim is None:
            embed_dim = self.embedding_size

        return nn.Linear(embed_dim, self.num_classes, bias=False)


registry.register_model_config("mobilevit_v2_0_25", MobileViT_v2, config={"width_factor": 0.25})
registry.register_model_config("mobilevit_v2_0_5", MobileViT_v2, config={"width_factor": 0.5})
registry.register_model_config("mobilevit_v2_0_75", MobileViT_v2, config={"width_factor": 0.75})
registry.register_model_config("mobilevit_v2_1_0", MobileViT_v2, config={"width_factor": 1.0})
registry.register_model_config("mobilevit_v2_1_25", MobileViT_v2, config={"width_factor": 1.25})
registry.register_model_config("mobilevit_v2_1_5", MobileViT_v2, config={"width_factor": 1.5})
registry.register_model_config("mobilevit_v2_1_75", MobileViT_v2, config={"width_factor": 1.75})
registry.register_model_config("mobilevit_v2_2_0", MobileViT_v2, config={"width_factor": 2.0})

registry.register_weights(
    "mobilevit_v2_1_0_il-common",
    {
        "description": "MobileViT v2 with width multiplier of 1.0 trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 17.6,
                "sha256": "2b45b7f2ffe3dd129d9a7e9690d2dfd0f93ac60f24d118b920a51bcb950fd95e",
            }
        },
        "net": {"network": "mobilevit_v2_1_0", "tag": "il-common"},
    },
)
registry.register_weights(
    "mobilevit_v2_1_5_il-common",
    {
        "description": "MobileViT v2 with width multiplier of 1.5 trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 38.8,
                "sha256": "acd28c3ee653b62c69ad765c1d99827cea5051deb6dbdd7b9c8d7612782c86a3",
            }
        },
        "net": {"network": "mobilevit_v2_1_5", "tag": "il-common"},
    },
)



"""
RepViT, adapted from
https://github.com/THU-MIG/RepViT/blob/main/model/repvit.py
and
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/repvit.py

Paper "RepViT: Revisiting Mobile CNN From ViT Perspective", https://arxiv.org/abs/2307.09283
"""

# Reference license: Apache-2.0 (both)

from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import SqueezeExcitation

from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import make_divisible


class RepConvBN(nn.Sequential):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        padding: tuple[int, int],
        groups: int,
        bn_weight_init: float,
        reparameterized: bool,
    ) -> None:
        super().__init__()
        self.c: nn.Module
        self.reparameterized = reparameterized
        self.add_module(
            "c",
            nn.Conv2d(
                in_channels,
                out_channels,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                groups=groups,
                bias=False,
            ),
        )
        if reparameterized is False:
            self.add_module("bn", nn.BatchNorm2d(out_channels))
            nn.init.zeros_(self.bn.bias)

        nn.init.constant_(self.bn.weight, bn_weight_init)

    def reparameterize(self) -> None:
        if self.reparameterized is True:
            return

        (c, bn) = self._modules.values()
        w = bn.weight / (bn.running_var + bn.eps) ** 0.5
        w = c.weight * w[:, None, None, None]
        b = bn.bias - bn.running_mean * bn.weight / (bn.running_var + bn.eps) ** 0.5
        self.c = nn.Conv2d(
            w.size(1) * self.c.groups,
            w.size(0),
            kernel_size=w.shape[2:],
            stride=self.c.stride,
            padding=self.c.padding,
            groups=self.c.groups,
            device=c.weight.device,
        )
        self.c.weight.data.copy_(w)
        self.c.bias.data.copy_(b)

        # Delete un-used branches
        for param in self.parameters():
            param.detach_()

        del self.bn

        self.reparameterized = True


class RepNormLinear(nn.Sequential):
    def __init__(self, in_dim: int, out_dim: int, reparameterized: bool) -> None:
        super().__init__()
        self.li: nn.Module
        self.reparameterized = reparameterized
        if reparameterized is False:
            self.add_module("bn", nn.BatchNorm1d(in_dim))

        self.add_module("li", nn.Linear(in_dim, out_dim))
        nn.init.trunc_normal_(self.li.weight, std=0.02)
        nn.init.zeros_(self.li.bias)

    def reparameterize(self) -> None:
        if self.reparameterized is True:
            return

        (bn, li) = self._modules.values()
        w = bn.weight / (bn.running_var + bn.eps) ** 0.5
        b = bn.bias - self.bn.running_mean * self.bn.weight / (bn.running_var + bn.eps) ** 0.5
        w = li.weight * w[None, :]
        if li.bias is None:
            b = b @ self.li.weight.T
        else:
            b = (li.weight @ b[:, None]).view(-1) + self.li.bias

        self.li = nn.Linear(w.size(1), w.size(0), device=li.weight.device)
        self.li.weight.data.copy_(w)
        self.li.bias.data.copy_(b)

        # Delete un-used branches
        for param in self.parameters():
            param.detach_()

        del self.bn

        self.reparameterized = True


class RepVitMLP(nn.Module):
    def __init__(self, in_dim: int, hidden_dim: int, reparameterized: bool):
        super().__init__()
        self.conv1 = RepConvBN(
            in_dim,
            hidden_dim,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            groups=1,
            bn_weight_init=1.0,
            reparameterized=reparameterized,
        )
        self.act = nn.GELU()
        self.conv2 = RepConvBN(
            hidden_dim,
            in_dim,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            groups=1,
            bn_weight_init=0.0,
            reparameterized=reparameterized,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.conv2(self.act(self.conv1(x)))


class RepVggDW(nn.Module):
    def __init__(self, dim: int, kernel_size: int, use_se: bool, reparameterized: bool) -> None:
        super().__init__()
        self.reparameterized = reparameterized
        self.conv = RepConvBN(
            dim,
            dim,
            kernel_size=(kernel_size, kernel_size),
            stride=(1, 1),
            padding=(kernel_size // 2, kernel_size // 2),
            groups=dim,
            bn_weight_init=1.0,
            reparameterized=reparameterized,
        )
        if reparameterized is False:
            self.conv1x1 = nn.Conv2d(dim, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), groups=dim)
            self.bn = nn.BatchNorm2d(dim)

        if use_se is True:
            self.se = SqueezeExcitation(dim, make_divisible(dim // 4, 8))
        else:
            self.se = nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.reparameterized is True:
            return self.se(self.conv(x))

        return self.se(self.bn(self.conv(x) + self.conv1x1(x) + x))

    def reparameterize(self) -> None:
        if self.reparameterized is True:
            return

        self.conv.reparameterize()
        conv = self.conv.c
        conv_w = conv.weight
        conv_b = conv.bias

        conv1x1 = self.conv1x1
        conv1x1_w = conv1x1.weight
        conv1x1_b = conv1x1.bias

        conv1x1_w = F.pad(conv1x1_w, [1, 1, 1, 1])

        identity = F.pad(
            torch.ones(conv1x1_w.shape[0], conv1x1_w.shape[1], 1, 1, device=conv1x1_w.device), [1, 1, 1, 1]
        )

        final_conv_w = conv_w + conv1x1_w + identity
        final_conv_b = conv_b + conv1x1_b

        conv.weight.data.copy_(final_conv_w)
        conv.bias.data.copy_(final_conv_b)

        bn = self.bn
        w = bn.weight / (bn.running_var + bn.eps) ** 0.5
        w = conv.weight * w[:, None, None, None]
        b = bn.bias + (conv.bias - bn.running_mean) * bn.weight / (bn.running_var + bn.eps) ** 0.5
        self.conv = nn.Conv2d(
            w.size(1) * self.conv.c.groups,
            w.size(0),
            kernel_size=w.shape[2:],
            stride=self.conv.c.stride,
            padding=self.conv.c.padding,
            groups=self.conv.c.groups,
            device=conv.weight.device,
        )
        self.conv.weight.data.copy_(w)
        self.conv.bias.data.copy_(b)

        # Delete un-used branches
        for param in self.parameters():
            param.detach_()

        del self.conv1x1
        del self.bn

        self.reparameterized = True


class RepViTBlock(nn.Module):
    def __init__(self, in_dim: int, mlp_ratio: float, kernel_size: int, use_se: bool, reparameterized: bool) -> None:
        super().__init__()

        self.token_mixer = RepVggDW(
            in_dim,
            kernel_size=kernel_size,
            use_se=use_se,
            reparameterized=reparameterized,
        )
        self.channel_mixer = RepVitMLP(in_dim, int(in_dim * mlp_ratio), reparameterized)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.token_mixer(x)

        identity = x
        x = self.channel_mixer(x)

        return identity + x


class RepVitDownsample(nn.Module):
    def __init__(self, in_dim: int, mlp_ratio: float, out_dim: int, kernel_size: int, reparameterized: bool) -> None:
        super().__init__()
        self.pre_block = RepViTBlock(in_dim, mlp_ratio, kernel_size, use_se=False, reparameterized=reparameterized)
        self.spatial_downsample = RepConvBN(
            in_dim,
            in_dim,
            kernel_size=(kernel_size, kernel_size),
            stride=(2, 2),
            padding=(kernel_size // 2, kernel_size // 2),
            groups=in_dim,
            bn_weight_init=1.0,
            reparameterized=reparameterized,
        )
        self.channel_downsample = RepConvBN(
            in_dim,
            out_dim,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            groups=1,
            bn_weight_init=1.0,
            reparameterized=reparameterized,
        )
        self.ffn = RepVitMLP(out_dim, int(out_dim * mlp_ratio), reparameterized=reparameterized)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.pre_block(x)
        x = self.spatial_downsample(x)
        x = self.channel_downsample(x)

        identity = x
        x = self.ffn(x)

        return x + identity


class RepViTStage(nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        depth: int,
        mlp_ratio: float,
        kernel_size: int,
        downsample: bool,
        reparameterized: bool,
    ) -> None:
        super().__init__()
        if downsample is True:
            self.downsample = RepVitDownsample(in_dim, mlp_ratio, out_dim, kernel_size, reparameterized)
        else:
            assert in_dim == out_dim
            self.downsample = nn.Identity()

        blocks = []
        use_se = True
        for _ in range(depth):
            blocks.append(RepViTBlock(out_dim, mlp_ratio, kernel_size, use_se, reparameterized))
            use_se = not use_se

        self.blocks = nn.Sequential(*blocks)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.downsample(x)
        x = self.blocks(x)

        return x


class RepViT(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        self.reparameterized = False
        embed_dims: list[int] = self.config["embed_dims"]
        depths: list[int] = self.config["depths"]

        self.stem = nn.Sequential(
            RepConvBN(
                in_channels=self.input_channels,
                out_channels=embed_dims[0] // 2,
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=(1, 1),
                groups=1,
                bn_weight_init=1.0,
                reparameterized=self.reparameterized,
            ),
            nn.GELU(),
            RepConvBN(
                in_channels=embed_dims[0] // 2,
                out_channels=embed_dims[0],
                kernel_size=(3, 3),
                stride=(2, 2),
                padding=(1, 1),
                groups=1,
                bn_weight_init=1.0,
                reparameterized=self.reparameterized,
            ),
        )

        num_stages = len(depths)
        prev_dim = embed_dims[0]

        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for idx in range(num_stages):
            stages[f"stage{idx+1}"] = RepViTStage(
                prev_dim,
                embed_dims[idx],
                depth=depths[idx],
                mlp_ratio=2.0,
                kernel_size=3,
                downsample=idx > 0,
                reparameterized=self.reparameterized,
            )
            return_channels.append(embed_dims[idx])
            prev_dim = embed_dims[idx]

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = embed_dims[-1]
        self.dist_classifier = self.create_classifier()
        self.classifier = self.create_classifier()
        self.distillation_output = False

    def reset_classifier(self, num_classes: int) -> None:
        self.num_classes = num_classes
        self.dist_classifier = self.create_classifier()
        self.classifier = self.create_classifier()

    def freeze(self, freeze_classifier: bool = True, unfreeze_features: bool = False) -> None:
        for param in self.parameters():
            param.requires_grad = False

        if freeze_classifier is False:
            for param in self.classifier.parameters():
                param.requires_grad = True

            for param in self.dist_classifier.parameters():
                param.requires_grad = True

        if unfreeze_features is True:
            for param in self.features.parameters():
                param.requires_grad = True

    def transform_to_backbone(self) -> None:
        self.features = nn.Identity()
        self.classifier = nn.Identity()
        self.dist_classifier = nn.Identity()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def create_classifier(self, embed_dim: Optional[int] = None) -> nn.Module:
        if self.num_classes == 0:
            return nn.Identity()

        if embed_dim is None:
            embed_dim = self.embedding_size

        return RepNormLinear(embed_dim, self.num_classes, self.reparameterized)

    def set_distillation_output(self, enable: bool = True) -> None:
        self.distillation_output = enable

    def classify(self, x: torch.Tensor) -> torch.Tensor:
        x_cls = self.classifier(x)
        x_dist = self.dist_classifier(x)

        if self.training is True and self.distillation_output is True:
            x = torch.stack([x_cls, x_dist], dim=1)
        else:
            x = (x_cls + x_dist) / 2

        return x

    def reparameterize_model(self) -> None:
        for module in self.modules():
            if hasattr(module, "reparameterize") is True:
                module.reparameterize()

        self.reparameterized = True


registry.register_model_config("repvit_m0_6", RepViT, config={"embed_dims": [48, 80, 160, 320], "depths": [1, 2, 9, 1]})
registry.register_model_config(
    "repvit_m0_9", RepViT, config={"embed_dims": [48, 96, 192, 384], "depths": [2, 2, 14, 2]}
)
registry.register_model_config(
    "repvit_m1_0", RepViT, config={"embed_dims": [56, 112, 224, 448], "depths": [2, 2, 14, 2]}
)
registry.register_model_config(
    "repvit_m1_1", RepViT, config={"embed_dims": [64, 128, 256, 512], "depths": [2, 2, 12, 2]}
)
registry.register_model_config(
    "repvit_m1_5", RepViT, config={"embed_dims": [64, 128, 256, 512], "depths": [4, 4, 24, 4]}
)
registry.register_model_config(
    "repvit_m2_3", RepViT, config={"embed_dims": [80, 160, 320, 640], "depths": [6, 6, 34, 2]}
)

registry.register_weights(
    "repvit_m0_6_il-common",
    {
        "description": "RepViT M0.6 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 10.1,
                "sha256": "4dfbad6d0f0b859d2d7bde9065dd6a93a72ff2fe32c923c7179004261bd2d700",
            }
        },
        "net": {"network": "repvit_m0_6", "tag": "il-common"},
    },
)
